#!/usr/bin/env python

#****************************************************************************
# flycore.py, provides non-GUI base classes for data
#
# FlyWay, a VFR/IFR Route Planner for Pilots
# Copyright (C) 2002, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#*****************************************************************************


from searchstr import SearchStr
from option import Option
from units import *
import time, math, sys, os.path

class Route:
    "Container class for waypoints and legs"
    def __init__(self, option):
        self.option = option
        self.wpList = []
        self.wpErrorList = []
        self.legList = []
        self.fltPlan = None
        self.mapData = None

    def findWp(self, keyWords):
        "Find waypoint by ID then search, return list"
        srchStr = keyWords.upper().strip()
        results = []
        # try to find by ID only for simple search string:
        if ' ' not in srchStr and '*' not in srchStr:
            if 3 <= len(srchStr) <= 4:
                results.append(Airport().setId(srchStr))
            if len(srchStr) <= 3:
                results.append(Navaid().setId(srchStr))
            # fix set id not req'd - will be found in search
            results = [wp for wp in results if wp]
            if results and len(srchStr) == 3:
                # return id matches only for common id length
                return results
        results = []    # avoid duplicate matches
        results.extend(Airport().search(srchStr))
        results.extend(Navaid().search(srchStr))
        results.extend(Fixpoint().search(srchStr))
        return results

    def readFile(self, fileName):
        "Read file contents into route, return 1/0 on success/fail"
        try:
            f = open(fileName, 'r')
            lines = [line.strip() for line in f.readlines()]
            f.close()
        except IOError:
            print 'Error - can not read file', fileName
            return 0
        try:
            self.fltPlan = None
            brkPos = lines.index('____')
            fltPlnList = lines[brkPos + 1:]
            self.fltPlan = PlanData(self.option, fltPlnList)
            lines = [line for line in lines[:brkPos] if line]
        except ValueError:
            pass
        typeDict = {'A': Airport, 'N': Navaid, 'F': Fixpoint}
        self.wpList = []
        self.wpErrorList = []
        for line in lines:
            wp = typeDict.get(line[0], Airport)().setId(line[2:])
            if wp:
                self.wpList.append(wp)
            else:
                self.wpErrorList.append(line[2:])
        return 1

    def writeFile(self, fileName):
        "Write route contents to file, return 1/0 on success/fail"
        try:
            f = open(fileName, 'w')
            f.writelines(['%s,%s\n' % (wp.wpType, wp.ident()) for wp in \
                          self.wpList])
            if self.fltPlan:
                f.write('____\n')
                f.writelines(self.fltPlan.writeFile())
            f.close()
        except IOError:
            print 'Error - can not write to file', fileName
            return 0
        return 1

    def updateLegs(self):
        "Replace legs for route"
        self.legList = []
        for i in range(len(self.wpList) - 1):
            self.legList.append(RouteLeg(self.wpList[i], self.wpList[i + 1], \
                                         self.option, self.legList and \
                                         self.legList[-1] or None))


class Waypoint:
    "Parent class for airport, navaid and fix data"
    id = 0
    latitude = 1
    longitude = 2
    magVar = 3
    stateName = 4
    stateAbbr = 5
    name = 6           # Airport & Navaid only
    aptAltName = navType = 7
    aptCity = navFreq = 8
    elevation = 9       # Airport & Navaid only
    aptFuelAvail = 10
    aptAwos = 11
    aptCtaf = 12
    aptTower = 13
    aptGround = 14
    rwy1dir1 = 15
    rwy1dir2 = 16
    rwy1length = 17
    rwy1width = 18
    rwy1surface = 19
    rwyToNext = 5
    aptFreqTypes = (('Tower', aptTower), ('Ground', aptGround), \
                    ('CTAF/UNIC', aptCtaf), ('ATIS/AWOS', aptAwos))
    def __init__(self, dataLine=None):
        self.dataList = []
        if dataLine:
            self.dataList = dataLine.replace('\n', '').split('\t')
            for field in self.__class__.lowCaseFields:
                self.dataList[field] = self.dataList[field].title()

    def setId(self, id):
        "Read data from buffer for ID, return self/None on pass/fail"
        if self.loadDataBuffer():
            srchStr = id + '\t'
            for line in self.__class__.dataBuffer:
                if line.startswith(srchStr):
                    self.dataList = self.__class__(line).dataList
                    return self
        return None

    def search(self, keyWords):
        "Search waypoint type for key words, return list"
        if self.loadDataBuffer():
            srch = SearchStr(keyWords)
            results = [self.__class__(line) for line in \
                       self.__class__.dataBuffer if \
                       srch.isMatch(line)]
            # take only matches to search fields
            results =  [wp for wp in results if \
                        srch.isMatch('\t'.join([wp.dataList[field] for \
                                                field in \
                                                self.__class__.searchFields])\
                                     .upper())]
            if results:
                self.dataList = results[0].dataList
            return results
        return []        

    def loadDataBuffer(self):
        "Load data into buffer if reqd, return 1 on success"
        if self.__class__.dataBuffer:
            return 1
        try:
            f = open(os.path.join(os.path.abspath(sys.path[0]), \
                                  self.__class__.fileName), 'r')
            self.__class__.dataBuffer = f.readlines()[1:]  # remove title line
            f.close()
            return 1
        except IOError:
            print 'Error - can not read data file', self.fileName
            return 0

    def getAllWaypoints(self):
        "Return a long list of all waypoints of this type"
        if self.loadDataBuffer():
            return [self.__class__(line) for line in \
                    self.__class__.dataBuffer if line.strip()]
        return []

    def latitudeValue(self):
        "Return latitude as a float"
        return float(self.dataList[Waypoint.latitude])

    def longitudeValue(self):
        "Return longitude as a float"
        return float(self.dataList[Waypoint.longitude])

    def magVarValue(self):
        "Return magnetic variation as a float"
        return float(self.dataList[Waypoint.magVar])

    def wpTypeChar(self):
        "Return first letter of type name"
        return self.__class__.wpType

    def elevationString(self):
        "Return elevation string"
        elev = self.dataList[Waypoint.elevation]
        if elev and elev != 'U':
            return elev + ' ft'
        return 'N/A'

    def magVarString(self):
        "Return magnetic variation string"
        if self.magVarValue() >= 0:
            return '%02.0f%s W' % (self.magVarValue(), chr(176))
        return '%02.0f%s E' % (-self.magVarValue(), chr(176))

    def mainFrequency(self):
        "Return primary frequency, overridden by types"
        return ''

    def ident(self):
        "Return id string"
        return self.dataList[Waypoint.id]

    def freqDescription(self):
        "Return string with frequency and wp description"
        freq = self.mainFrequency()
        if freq:
            return '(%s)  %s' % (freq, self.description())
        return self.description()

    def detailNumLines(self):
        "Return num lines in detail list, 2 used for wp, descr, elev, magvar"
        return self.__class__.detailLines


class Airport(Waypoint):
    "Waypoint class for airport data"
    fileName = 'apt.dat'
    wpType = 'A'
    dataBuffer = []
    searchFields = [Waypoint.id, Waypoint.stateName, Waypoint.stateAbbr, \
                    Waypoint.name, Waypoint.aptAltName, Waypoint.aptCity]
    lowCaseFields = [Waypoint.stateName, Waypoint.name, \
                     Waypoint.aptAltName, Waypoint.aptCity]
    rwySurfDict = {'ASP': 'Asphalt', 'CON': 'Concrete', 'GRE': 'Graded', \
                   'GRS': 'Grass', 'GVL': 'Gravel', 'BRI': 'Misc hard', \
                   'COP': 'Misc hard', 'PEM': 'Misc hard', 'PER': 'Misc hard',\
                   'BIT': 'Misc soft', 'CLA': 'Misc soft', 'COM': 'Misc soft',\
                   'COR': 'Misc soft', 'ICE': 'Misc soft', 'LAT': 'Misc soft',\
                   'MAC': 'Misc soft', 'MEM': 'Misc soft', 'MIX': 'Misc soft',\
                   'PSP': 'Misc soft', 'SAN': 'Misc soft', 'SNO': 'Misc soft'}

    def __init__(self, dataLine=None):
        Waypoint.__init__(self, dataLine)

    def description(self):
        "Return string description based on several fields"
        result = self.dataList[Waypoint.name]
        if self.dataList[Waypoint.aptAltName]:
            result = '%s / %s' % (result, self.dataList[Waypoint.aptAltName])
        result += ' Airport'
        if self.dataList[Waypoint.aptCity]:
            result = '%s, %s' % (result, self.dataList[Waypoint.aptCity])
        return '%s, %s' % (result, self.dataList[Waypoint.stateName])

    def mainFrequency(self):
        "Return primary frequency"
        if self.dataList[Waypoint.aptTower]:
            return self.freqAtPos(Waypoint.aptTower)
        return self.freqAtPos(Waypoint.aptCtaf)

    def freqAtPos(self, pos):
        "Return given frequency"
        if pos == Waypoint.aptCtaf and \
           self.dataList[pos] == self.dataList[Waypoint.aptTower]:
            return ''
        if self.dataList[pos]:
            return '%03.2f MHz' % float(self.dataList[pos])
        return ''

    def fuelAvail(self):
        "Return fuel string"
        if self.dataList[Waypoint.aptFuelAvail] == 'Y':
            return 'Fuel is Available'
        return 'Fuel Not Available'

    def rwyName(self, num):
        "Return string with ID numbers for runway num 0-5"
        delta = num * Waypoint.rwyToNext
        if len(self.dataList) > Waypoint.rwy1dir1 + delta and \
           self.dataList[Waypoint.rwy1dir1 + delta]:
            return 'Rwy %s/%s' % (self.dataList[Waypoint.rwy1dir1 + delta], \
                                   self.dataList[Waypoint.rwy1dir2 + delta])
        return ''

    def rwyDesc(self, num):
        "Return string with info for runway num 0-5"
        delta = num * Waypoint.rwyToNext
        return '%sx%s ft %s' % (self.dataList[Waypoint.rwy1length + delta], \
                                self.dataList[Waypoint.rwy1width + delta], \
                                Airport.rwySurfDict.get(self.dataList\
                                        [Waypoint.rwy1surface + delta], ''))

    def detailList(self):
        "Return list of fields with all airport details"
        freqList = ['%s %s' % (self.freqAtPos(type[1]), type[0]) for type \
                    in Waypoint.aptFreqTypes if self.freqAtPos(type[1])]
        rwyList = ['%s:  %s' % (self.rwyName(num), self.rwyDesc(num)) for \
                   num in range(6) if self.rwyName(num)]
        list = ['%s Elevation' % self.elevationString(), \
                '%s Magn. Var.' % self.magVarString(), self.fuelAvail()]
        for couple in map(None, freqList, rwyList):
            list.extend(couple)
        return [item and item or '' for item in list]  # remove None's

    def detailNumLines(self):
        "Return num lines in detail list, 2 used for wp, descr, elev, magvar"
        freqList = [type for type in Waypoint.aptFreqTypes if \
                    self.freqAtPos(type[1])]
        rwyList = [num for num in range(6) if self.rwyName(num)]
        return max(len(freqList), len(rwyList)) + 2


class Navaid(Waypoint):
    "Waypoint class for navaid data"
    fileName = 'nav.dat'
    wpType = 'N'
    dataBuffer = []
    detailLines = 3
    searchFields = [Waypoint.id, Waypoint.stateName, \
                    Waypoint.stateAbbr, Waypoint.name]
    lowCaseFields = [Waypoint.stateName, Waypoint.name]

    def __init__(self, dataLine=None):
        Waypoint.__init__(self, dataLine)

    def description(self):
        "Return string description based on several fields"
        return '%s %s, %s' % (self.dataList[Waypoint.name], \
                              self.dataList[Waypoint.navType], \
                              self.dataList[Waypoint.stateName])

    def mainFrequency(self):
        "Return primary frequency"
        if self.dataList[Waypoint.navType].find('NDB') >= 0:
            return '%03.2f kHz' % float(self.dataList[Waypoint.navFreq])
        return '%03.2f MHz' % float(self.dataList[Waypoint.navFreq])

    def detailList(self):
        "Return list of fields with all navaid details"
        return ['%s Elevation' % self.elevationString(), \
                '%s Magn. Var.' % self.magVarString(), '', \
                '%s Frequency' % self.mainFrequency(), '']


class Fixpoint(Waypoint):
    "Waypoint class for fix data"
    fileName = 'fix.dat'
    wpType = 'F'
    dataBuffer = []
    detailLines = 2
    searchFields = [Waypoint.id, Waypoint.stateName, Waypoint.stateAbbr]
    lowCaseFields = [Waypoint.stateName]

    def __init__(self, dataLine=None):
        Waypoint.__init__(self, dataLine)

    def description(self):
        "Return string description based on several fields"
        return 'Intersection, %s' % self.dataList[Waypoint.stateName]

    def detailList(self):
        "Return list of fields with all fixpoint details"
        return ['', '%s Magn. Var.' % self.magVarString(), '']


class RouteLeg:
    "Calculates parameters between two waypoints"
    nmLat = 60.00721
    nmLng = 60.10793
    maxAirspeed = 9999
    maxWindSpeed = 999
    def __init__(self, fromWp, toWp, option, prevLeg=None):
        yDistance = (toWp.latitudeValue() - fromWp.latitudeValue()) \
                    * RouteLeg.nmLat
        xDistance = (math.cos(fromWp.latitudeValue() * math.pi/180) \
                     + math.cos(toWp.latitudeValue() * math.pi/180)) \
                     * (toWp.longitudeValue() - fromWp.longitudeValue()) \
                     * (RouteLeg.nmLng / 2)
        self.distance = math.sqrt(yDistance**2 + xDistance**2)
        if yDistance != 0:
            self.absCourse = Angle(math.atan(xDistance/yDistance)*180/math.pi)
            if yDistance < 0:
                self.absCourse = Angle(self.absCourse - 180.0)
        elif xDistance >= 0:
            self.absCourse = Angle(90.0)
        else:
            self.absCourse = Angle(-90.0)

        airspeed = option.intData('Airspeed', 1, RouteLeg.maxAirspeed)
        windSpeed = option.intData('WindSpeed', 0, RouteLeg.maxWindSpeed)
        windDir = Angle(option.intData('WindDir', 0, 360))
        self.is12hrClock = option.boolData('12hrClock')

        absHeading = Angle(self.absCourse - (180/math.pi) \
                           * math.asin(float(windSpeed) / airspeed) \
                           * math.sin((self.absCourse - windDir) \
                                      * math.pi/180))
        self.groundSpeed = math.cos((absHeading - self.absCourse) \
                                    * math.pi/180) * airspeed \
                           + windSpeed * math.cos((windDir - self.absCourse \
                                                   + 180.0) * math.pi/180)
        magVar = (fromWp.magVarValue() + toWp.magVarValue()) / 2
        self.course = Angle(self.absCourse + magVar)
        self.heading = Angle(absHeading + magVar)
        self.ete = TimeSpan(self.distance / self.groundSpeed)
        if prevLeg:
            self.eta = Time(prevLeg.eta + self.ete)
            self.totalDist = prevLeg.totalDist + self.distance
            self.totalTime = TimeSpan(prevLeg.totalTime + self.ete)
        else:
            self.eta = Time(option.intData('DepartHour', 0, 23) \
                            + option.intData('DepartMinute', 0, 59) / 60.0 \
                            + self.ete)
            self.totalDist = self.distance
            self.totalTime = self.ete

    def output(self):
        "Return tuple of 6 result strings"
        return (self.course.angleStr(), self.heading.angleStr(), \
                '%3.0f nmi' % self.distance, '%3.0f kts' % self.groundSpeed, \
                self.ete.timeStr(), self.eta.timeStr(not self.is12hrClock))

    def outputTotals(self):
        "Return tuple of 6 strings with total distance and time"
        return ('', '', '%3.0f nmi' % self.totalDist, '', \
                self.totalTime.timeStr(), '')


class PlanData:
    "Stores flight plan data and sets defaults"
    type1 = 0
    type2 = 1
    type3 = 2
    craftID = 3
    equip = 4
    altitude = 5
    route1 = 6
    route2 = 7
    route3 = 8
    route4 = 9
    remark1 = 10
    remark2 = 11
    remark3 = 12
    fuelHr = 13
    fuelMin = 14
    alternate = 15
    pilot1 = 16
    pilot2 = 17
    pilot3 = 18
    numAboard = 19
    color = 20
    listLength = 21
    intData = ((altitude, 0, 60000), (fuelHr, 0, 24), \
               (fuelMin, 0, 59), (numAboard, 1, 99))
    optionData = ('PlanType1', 'PlanType2', 'PlanType3', 'PlanCraftID', \
                  'PlanEquip', 'PlanAltitude', '', '', '', '', \
                  'PlanRemark1', 'PlanRemark2', 'PlanRemark3', \
                  'PlanFuelHr', 'PlanFuelMin', '', 'PlanPilot1', \
                  'PlanPilot2', 'PlanPilot3', 'PlanNumAboard', 'PlanColor')
    def __init__(self, option, list=None):
        self.data = [''] * PlanData.listLength
        self.option = option
        self.intList = [data[0] for data in PlanData.intData]
        self.intMinList = [None] * PlanData.listLength
        self.intMaxList = [None] * PlanData.listLength
        for data in PlanData.intData:
            self.intMinList[data[0]] = data[1]
            self.intMaxList[data[0]] = data[2]
        if list:
            self.readFile(list)

    def readFile(self, list):
        "Set plan data to list from open file"
        self.data = list
        self.data.extend([''] * (PlanData.listLength - len(self.data)))

    def writeFile(self):
        "Return list to write to a file"
        return [line + '\n' for line in self.data]

    def setToDefaults(self):
        "Set this plan to all current option settings"
        for i in range(PlanData.listLength):
            if PlanData.optionData[i]:
                if i in self.intList:
                    self.data[i] = str(self.option.\
                                       intData(PlanData.optionData[i], \
                                               self.intMinList[i], \
                                               self.intMaxList[i]))
                else:
                    self.data[i] = self.option.strData(PlanData.\
                                                       optionData[i], 1)

    def setDefaultRoute(self, route):
        "Add direct routing based on route"
        self.data[PlanData.route1] = ' .. '.join([wp.ident() for wp in \
                                                  route.wpList[1:-1]])
        self.data[PlanData.route2] = self.data[PlanData.route3] = \
                                     self.data[PlanData.route4] = ''
        self.data[PlanData.alternate] = ''

    def updateDefaults(self):
        "Update options with entries from this plan"
        for i in range(PlanData.listLength):
            if PlanData.optionData[i]:
                self.option.changeData(PlanData.optionData[i], self.data[i], 1)
        self.option.writeChanges()


class MapData:
    "Calculates and stores mapping info"
    marginFactor = 1.1
    mapPixels = 500
    minMapSize = 20
    def __init__(self, route):
        self.route = route
        self.routeList = []
        self.airportList = []
        self.navaidList = []
        self.fixList = []
        self.maxLat = 0.0
        self.minLat = 0.0
        self.maxLng = 0.0
        self.minLng = 0.0
        self.calcBounds()
        self.addRouteWp()

    def calcBounds(self):
        "Determine map boundaries for a square map"
        if self.route.wpList:
            self.maxLat = max([wp.latitudeValue() for wp in self.route.wpList])
            self.minLat = min([wp.latitudeValue() for wp in self.route.wpList])
            self.maxLng = max([abs(wp.longitudeValue()) for wp \
                               in self.route.wpList])
            self.minLng = min([abs(wp.longitudeValue()) for wp \
                               in self.route.wpList])
            lngFactor = (math.cos(self.minLat * math.pi/180) \
                         + math.cos(self.maxLat * math.pi/180)) / 2.0
            xSize = (self.maxLng - self.minLng) * RouteLeg.nmLng * lngFactor
            ySize = (self.maxLat - self.minLat) * RouteLeg.nmLat
            if xSize == 0 and ySize == 0:
                self.maxLat += MapData.minMapSize / (2 * RouteLeg.nmLat)
                self.minLat -= MapData.minMapSize / (2 * RouteLeg.nmLat)
                self.maxLng += MapData.minMapSize \
                               / (2 * RouteLeg.nmLng * lngFactor)
                self.minLng -= MapData.minMapSize \
                               / (2 * RouteLeg.nmLng * lngFactor)
            elif xSize > ySize:     # expand short side to make map square
                delta = (xSize - ySize) / (2 * RouteLeg.nmLat)
                self.maxLat += delta
                self.minLat -= delta
            else:
                delta = (ySize - xSize) / (2 * RouteLeg.nmLng * lngFactor)
                self.maxLng += delta
                self.minLng -= delta
            self.expandBounds(MapData.marginFactor)

    def expandBounds(self, factor):
        "Enlarge boundary by factor on each side"
        latDelta = (self.maxLat - self.minLat) * (factor - 1)
        self.maxLat += latDelta
        self.minLat -= latDelta
        lngDelta = (self.maxLng - self.minLng) * (factor - 1)
        self.maxLng += lngDelta
        self.minLng -= lngDelta

    def addRouteWp(self):
        "Add waypoints from the route to the map list"
        self.routeList = [MapItem(wp, self) for wp in self.route.wpList]
        if self.route.legList:
            angList = [Angle(self.route.legList[0].absCourse + 180.0)]
            prevAng = self.route.legList[0].absCourse
            for leg in self.route.legList[1:]:
                newAng = Angle(leg.absCourse + 180.0)
                angList.append((prevAng + newAng) / 2.0)  # average angles
                if abs(prevAng - newAng) > 180:
                    angList[-1] = Angle(angList[-1] + 180.0)
                prevAng = leg.absCourse
            angList.append(prevAng)
            octantList = [int(round(ang / 45.0)) for ang in angList]
            for i in range(len(self.routeList)):
                self.routeList[i].labelOctant = octantList[i]
                if octantList[i] == 8:
                    self.routeList[i].labelOctant = 0

    def addAirports(self):
        "Add additional airports from the database to the map list"
        self.airportList = [MapItem(wp, self) for wp in \
                            Airport().getAllWaypoints() if self.isOnMap(wp)]

    def addNavaids(self):
        "Add additional navaids from the database to the map list"
        self.navaidList = [MapItem(wp, self) for wp in \
                           Navaid().getAllWaypoints() if self.isOnMap(wp)]

    def addFixes(self):
        "Add additional fixes from the database to the map list"
        self.fixList = [MapItem(wp, self) for wp in \
                        Fixpoint().getAllWaypoints() if self.isOnMap(wp)]

    def isOnMap(self, wp):
        "Return 1 if wp is within map bounds"
        return self.minLat <= wp.latitudeValue() <= self.maxLat \
               and self.minLng <= abs(wp.longitudeValue()) <= self.maxLng


class MapItem:
    "Stores data for an item included on the map"
    def __init__(self, waypoint, mapData):
        self.waypoint = waypoint
        self.xCoord = MapData.mapPixels - \
                      int((abs(self.waypoint.longitudeValue()) \
                           - mapData.minLng) \
                          / (mapData.maxLng - mapData.minLng) \
                          * MapData.mapPixels)
        self.yCoord = MapData.mapPixels - \
                      int((self.waypoint.latitudeValue() - mapData.minLat) \
                          / (mapData.maxLat - mapData.minLat) \
                          * MapData.mapPixels)
        self.labelOctant = 4        # default label position for route labels


if __name__ == '__main__':
    option = Option('flyway', 20)
    option.loadAll(['Airspeed            100'])
    option.addData('WindSpeed', '0')
    option.addData('WindDir', '360')
    option.addData('DepartHour', '8')
    option.addData('DepartMinute', '0')
    rt = Route()
    if rt.readFile('afj-fdk.fly'):
        for wp in rt.wpList:
            print wp.ident(), wp.description()
        rt.updateLegs(option)
        for leg in rt.legList:
            print '   '.join(leg.output())
