#!/usr/bin/env python

#****************************************************************************
# flyway.py, the main program file, version 0.2.2
#
# FlyWay, a VFR/IFR Route Planner for Pilots
# Copyright (C) 2002, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the 
# terms of the GNU General Public License, Version 2.  This program is 
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#*****************************************************************************

from flycore import *
from option import Option
from optiondlg import *
from textpaint import *
from tmpcontrol import *
from printpreview import *
from helpview import *
from recentfiles import RecentFiles
from qt import *
from xpm import *
import sys, getopt, os.path

class RouteView(QScrollView):
    "Main route window"
    wpStart = (50, 70)
    descStart = 80
    colStart = 130
    colSize = 68
    rowSize = 22
    totalWidth = 510
    margin = 10
    paramStart = 15
    editStart = 30
    headings = ('Course', 'Heading', 'Distance', 'Grnd Spd', 'ETE', 'ETA')
    def __init__(self, parent=None, name=None):
        QScrollView.__init__(self, parent, name)
        self.viewport().setBackgroundMode(QWidget.PaletteBase)
        self.viewport().setFocusPolicy(QWidget.StrongFocus)
        self.numSelect = 0
        self.tmpEdit = None
        self.normFont = self.font()
        self.ulFont = self.font()
        self.ulFont.setUnderline(1)
        self.boldFont = self.font()
        self.boldFont.setWeight(QFont.Bold)
        self.smFont = self.font()
        self.smFont.setPointSize(self.font().pointSize() - 2)
        self.lgFont = QFont('helvetica', self.font().pointSize() + 8)
        self.recalc()

    def recalc(self, updateViews=1):
        "Update legs & view size & repaint"
        FlyMainWin.route.updateLegs()
        self.resizeContents(RouteView.totalWidth, RouteView.wpStart[1] \
                            + len(FlyMainWin.route.wpList) \
                            * RouteView.rowSize * 2)
        self.airspeedStr = 'Airspeed: %d kts' \
                           % FlyMainWin.option.intData('Airspeed', 1, \
                                                       RouteLeg.maxAirspeed)
        windSpeed = FlyMainWin.option.intData('WindSpeed', 0, \
                                              RouteLeg.maxWindSpeed)
        if windSpeed:
            self.windStr = 'Wind: %s at %d kts' % \
                                (Angle(FlyMainWin.option.intData('WindDir', \
                                                                 0, 360)).\
                                 angleStr(), windSpeed)
        else:
            self.windStr = 'Wind: none'
        is12hrClock = FlyMainWin.option.boolData('12hrClock')
        self.departStr = 'Departure Time: %s' % \
                         Time(FlyMainWin.option.intData('DepartHour', 0, 23) \
                           + FlyMainWin.option.intData('DepartMinute', 0, 59) \
                              / 60.0).timeStr(not is12hrClock, is12hrClock)
        self.setSelectWp(self.numSelect)
        self.viewport().update()
        if updateViews:
            self.emit(PYSIGNAL('recalcDone'), ())

    def paintTable(self, painter, startRow=None, stopRow=None):
        "Paint table of parameters, waypoints and legs"
        if startRow == None:
            startRow = -1
        if stopRow == None:
            stopRow = len(FlyMainWin.route.wpList) * 2
        if startRow < 0:       # paint parameters and column headings
            painter.setFont(self.normFont)
            painter.drawText(RouteView.margin, RouteView.paramStart, \
                             self.airspeedStr)
            drawCenterText(painter, RouteView.totalWidth / 2, \
                           RouteView.paramStart, self.windStr)
            drawRightText(painter, RouteView.totalWidth - RouteView.margin, \
                          RouteView.paramStart, self.departStr)
            painter.drawLine(RouteView.margin, \
                             RouteView.paramStart + 5, \
                             RouteView.totalWidth - RouteView.margin, \
                             RouteView.paramStart + 5)
            painter.setFont(self.ulFont)
            self.__drawRow(painter, RouteView.colStart, \
                           RouteView.wpStart[1] - RouteView.rowSize, \
                           RouteView.headings)
        y = RouteView.wpStart[1]
        startRow = startRow * RouteView.rowSize + RouteView.wpStart[1]
        stopRow = stopRow * RouteView.rowSize + RouteView.wpStart[1]
        for wp in FlyMainWin.route.wpList:        # paint waypoints
            if startRow <= y < stopRow:
                painter.setFont(self.boldFont)
                drawCenterText(painter, RouteView.wpStart[0], y, wp.ident()) 
                painter.setFont(self.smFont)
                painter.drawText(RouteView.descStart, y, wp.freqDescription())
            y += 2 * RouteView.rowSize
        painter.setFont(self.normFont)
        y = RouteView.wpStart[1] + RouteView.rowSize
        for leg in FlyMainWin.route.legList:      # paint legs
            if startRow <= y < stopRow:
                self.__drawRow(painter, RouteView.colStart, y, leg.output())
            y += 2 * RouteView.rowSize
        if len(FlyMainWin.route.legList) > 1 and startRow <= y < stopRow:
            painter.setFont(self.ulFont)          # paint totals
            painter.drawText(RouteView.descStart, y, \
                             'Totals:')
            painter.setFont(self.normFont)
            self.__drawRow(painter, RouteView.colStart, y, leg.outputTotals())

    def drawContents(self, painter, clipX=None, clipY=None, \
                     clipW=None, clipH=None):
        "Paint view"
        if clipX == None:   # none is for wrong overridden function
            return QFrame.drawContents(self, painter)
        pal = self.palette()
        painter.setPen(pal.color(QPalette.Active, QColorGroup.Text))
        self.paintTable(painter)
        if FlyMainWin.route.wpList:
            painter.setFont(self.lgFont)
            painter.drawText(RouteView.margin, RouteView.wpStart[1] + 3 \
                             + RouteView.rowSize * 2 * self.numSelect, \
                             chr(187))

    def __drawRow(self, painter, x, y, list):
        for text in list:
            drawCenterText(painter, x, y, text)
            x += RouteView.colSize

    def insertWaypoint(self):
        rowPos = RouteView.wpStart[1] + RouteView.rowSize * 2 * self.numSelect
        self.ensureVisible(RouteView.wpStart[0], rowPos, 12, 12)
        self.tmpEdit = TmpEdit('', self.viewport())
        pos = self.contentsToViewport(RouteView.editStart, rowPos - 16)
        self.tmpEdit.setGeometry(pos[0], pos[1], RouteView.totalWidth \
                                 - RouteView.editStart - RouteView.margin, \
                                 RouteView.rowSize)
        self.tmpEdit.show()
        self.tmpEdit.setFocus()
        self.connect(self.tmpEdit, PYSIGNAL('editDone'), self.insertDone)
        FlyMainWin.toolBar.setEnabled(0)
        self.emit(PYSIGNAL('statusBarMsg'), \
                  ('Enter ID or search string', 1000000))

    def insertDone(self):
        "Called by tmpEdit after return or focus change"
        wp = None
        self.emit(PYSIGNAL('statusBarMsg'), ('', 1000))
        FlyMainWin.toolBar.setEnabled(1)
        text = str(self.tmpEdit.text())
        if text:
            results = FlyMainWin.route.findWp(text)
            if not results:        # none found
                self.tmpEdit.clear()
                self.tmpEdit.active = 1
                self.emit(PYSIGNAL('statusBarMsg'), \
                          ('Nothing found, try again', 1000000))
                return
            if len(results) == 1:  # one find
                wp = results[0]
            else:                  # multiple finds
                dlg = SearchDlg(results, text, self, None, 1)
                if dlg.exec_loop() == QDialog.Accepted:
                    wp = dlg.result()
        if FlyMainWin.route.wpList:
            del FlyMainWin.route.wpList[self.numSelect]
        if wp:
            FlyMainWin.route.wpList.insert(self.numSelect, wp)
            FlyMainWin.docModified = 1
        else:
            if self.numSelect:
                self.numSelect -= 1
        self.tmpEdit.close(1)
        self.tmpEdit = None
        self.setFocus()
        self.recalc()

    def editInBefore(self):
        "Add new waypoint before the selected one"
        if FlyMainWin.route.wpList:
            FlyMainWin.route.wpList.insert(self.numSelect, FlyMainWin.route.\
                                           wpList[self.numSelect])
            self.recalc(0)
        self.insertWaypoint()

    def editInAfter(self):
        "Add new waypoint after the selected one"
        if FlyMainWin.route.wpList:
            FlyMainWin.route.wpList.insert(self.numSelect + 1, \
                                           FlyMainWin.route.\
                                           wpList[self.numSelect])
            self.numSelect += 1
            self.recalc(0)
        self.insertWaypoint()

    def editDelete(self):
        "Delete selected waypoint"
        if FlyMainWin.route.wpList:
            del FlyMainWin.route.wpList[self.numSelect]
            if self.numSelect >= len(FlyMainWin.route.wpList) and \
               FlyMainWin.route.wpList:
                self.numSelect = len(FlyMainWin.route.wpList) - 1
            FlyMainWin.docModified = 1
            self.recalc()

    def editUp(self):
        "Move the selected waypoint upward"
        if self.numSelect > 0:
            FlyMainWin.route.wpList[self.numSelect], \
               FlyMainWin.route.wpList[self.numSelect - 1] = \
               FlyMainWin.route.wpList[self.numSelect - 1], \
               FlyMainWin.route.wpList[self.numSelect]
            self.numSelect = self.numSelect - 1
            FlyMainWin.docModified = 1
            self.recalc()

    def editDown(self):
        "Move the selected waypoint downward"
        if self.numSelect < len(FlyMainWin.route.wpList) - 1:
            FlyMainWin.route.wpList[self.numSelect], \
               FlyMainWin.route.wpList[self.numSelect + 1] = \
               FlyMainWin.route.wpList[self.numSelect + 1], \
               FlyMainWin.route.wpList[self.numSelect]
            self.numSelect = self.numSelect + 1
            FlyMainWin.docModified = 1
            self.recalc()

    def editReverse(self):
        "Reverse the order of all waypoints"
        if FlyMainWin.route.wpList:
            FlyMainWin.route.wpList.reverse()
            self.numSelect = 0
            FlyMainWin.docModified = 1
            self.recalc()

    def setSelectWp(self, num):
        "Set selected waypoint number"
        if num < len(FlyMainWin.route.wpList):
            self.ensureVisible(RouteView.wpStart[0], RouteView.wpStart[1] \
                               + RouteView.rowSize * 2 * num, 12, 12)
            self.numSelect = num
            self.emit(PYSIGNAL('selChange'), (FlyMainWin.route.wpList[num],))
        else:
            self.numSelect = 0
            self.emit(PYSIGNAL('selChange'), (None,))
        self.viewport().update()

    def contentsMousePressEvent(self, event):
        "Mouse waypoint selection control"
        line = (event.pos().y() - RouteView.wpStart[1] + 16) \
               / RouteView.rowSize
        if not line % 2 and 0 <= line / 2 < len(FlyMainWin.route.wpList):
            self.setSelectWp(line / 2)

    def contentsMouseReleaseEvent(self, event):
        "Mouse waypoint selection control"
        if event.button() == Qt.RightButton:
            line = (event.pos().y() - RouteView.wpStart[1] + 16) \
                   / RouteView.rowSize
            if not line % 2 and line / 2 == self.numSelect:
                pt = self.mapToGlobal(self.contentsToViewport(event.pos()))
                FlyMainWin.routePopup.popup(pt)

    def keyPressEvent(self, event):
        "Key bindings"
        if event.key() == Qt.Key_Up:
            if self.numSelect > 0:
                self.setSelectWp(self.numSelect - 1)
        elif event.key() == Qt.Key_Down:
            if self.numSelect < len(FlyMainWin.route.wpList) - 1:
                self.setSelectWp(self.numSelect + 1)
        elif event.key() == Qt.Key_PageUp:
            self.scrollBy(0, -self.visibleHeight() + RouteView.rowSize)
        elif event.key() == Qt.Key_PageDown:
            self.scrollBy(0, self.visibleHeight() - RouteView.rowSize)
        elif event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
            if FlyMainWin.option.boolData('AddOnEnter'):
                self.editInAfter()
        elif event.key() == Qt.Key_Delete:
            if FlyMainWin.option.boolData('DelKey'):
                self.editDelete()
        else:
            QWidget.keyPressEvent(self, event)


class WpView(QScrollView):
    "Waypoint description window"
    idStart = (20, 20)
    thirdsCol = (100, 250, 400)
    halfCol = (80, 280)
    rowSize = 20
    def __init__(self, parent=None, name=None):
        QScrollView.__init__(self, parent, name)
        self.viewport().setBackgroundMode(QWidget.PaletteBase)
        self.viewport().setFocusPolicy(QWidget.NoFocus)
        self.selWp = None
        self.normFont = self.font()

    def setSelWp(self, wp):
        "Slot to set the selected waypoint"
        self.selWp = wp
        if self.selWp:
            self.resizeContents(RouteView.totalWidth, WpView.idStart[1] \
                                + self.selWp.detailNumLines() * WpView.rowSize)
        else:
            self.resizeContents(1, 1)
        self.viewport().update()

    def paintWp(self, painter, wp):
        "Paint waypoint info"
        painter.setFont(self.normFont)
        painter.drawText(WpView.idStart[0], WpView.idStart[1], wp.ident())
        painter.drawText(WpView.halfCol[0], WpView.idStart[1], \
                         wp.description())
        info = wp.detailList()
        row = WpView.idStart[1] + WpView.rowSize
        for col in WpView.thirdsCol:
            painter.drawText(col, row, info.pop(0))
        while info:
            row += WpView.rowSize
            painter.drawText(WpView.halfCol[0], row, info.pop(0))
            painter.drawText(WpView.halfCol[1], row, info.pop(0))

    def drawContents(self, painter, clipX=None, clipY=None, \
                     clipW=None, clipH=None):
        "Paint view"
        if clipX == None:   # none is for wrong overridden function
            return QFrame.drawContents(self, painter)
        pal = self.palette()
        painter.setPen(pal.color(QPalette.Active, QColorGroup.Text))
        if self.selWp:
            self.paintWp(painter, self.selWp)


class PlanView(QScrollView):
    "Flight plan window"
    colPos = (0, 23, 46, 92, 126, 172, 218, 276, 366, 424, 482, 540)
    rowPos = (0, 17, 29, 41, 53, 65, 77, 150, 162, 174, 210, 222, 234, \
              270, 282, 305)
    vertLines = ((0, 0, 15), (1, 2, 5), (2, 1, 5), (2, 11, 13), (3, 10, 13), \
                 (4, 1, 5), (4, 7, 10), (4, 13, 15), (5, 8, 10), (6, 1, 5), \
                 (6, 7, 13), (7, 1, 5), (8, 1, 5), (9, 2, 5), (10, 1, 5), \
                 (10, 10, 13), (11, 0, 15))
    horizLines = ((0, 0, 11), (1, 0, 11), (2, 0, 2), (2, 8, 10), (3, 0, 2), \
                  (4, 0, 2), (5, 0, 11), (7, 0, 11), (8, 4, 6), (10, 0, 11), \
                  (11, 0, 3), (13, 0, 11), (15, 0, 11))
    viewMargin = 10
    def __init__(self, parent=None, name=None):
        QScrollView.__init__(self, parent, name)
        self.viewport().setBackgroundMode(QWidget.PaletteBase)
        self.viewport().setFocusPolicy(QWidget.StrongFocus)
        if sys.platform == 'win32':
            self.setCaption('Flight Plan  (PyQt)')
        else:
            self.setCaption('Flight Plan')
        self.setIcon(QPixmap(fly_xpm))
        self.resizeContents(PlanView.colPos[-1] + 2 * PlanView.viewMargin, \
                            PlanView.rowPos[-1] + 2 * PlanView.viewMargin)
        self.normFont = self.font()
        self.smFont = QFont('helvetica', 6)

        type1 = BoolEntry(self.rect(0, 2, 1, 3), PlanData.type1, 1, self)
        self.entries = [type1]
        self.selectedEntry = type1
        type2 = BoolEntry(self.rect(0, 3, 1, 4), PlanData.type2, 2, self)
        self.entries.append(type2)
        type3 = BoolEntry(self.rect(0, 4, 1, 5), PlanData.type3, 3, self)
        self.entries.append(type3)
        self.connect(type1, PYSIGNAL('checked'), type2.uncheck)
        self.connect(type2, PYSIGNAL('checked'), type3.uncheck)
        self.connect(type3, PYSIGNAL('checked'), type1.uncheck)
        self.entries.append(LineEntry(self.rect(2, 3, 4, 5), \
                                      PlanData.craftID, self))
        self.entries.append(LineEntry(self.rect(4, 3, 6, 5), \
                                      PlanData.equip, self))
        self.entries.append(SpeedEntry(self.rect(6, 3, 7, 5), None, self))
        self.entries.append(DepartEntry(self.rect(8, 3, 9, 5), None, self))
        alt = IntEntry(self.rect(10, 3, 11, 5), PlanData.altitude, self)
        alt.step = 500
        self.entries.append(alt)
        self.entries.append(MultiEntry(self.rect(0, 6, 11, 7), \
                                       PlanData.route1, 4, self))
        self.entries.append(MultiEntry(self.rect(6, 8, 11, 10), \
                                       PlanData.remark1, 3, self))
        self.entries.append(IntEntry(self.rect(0, 12, 2, 13), \
                                     PlanData.fuelHr, self))
        fuelMin = IntEntry(self.rect(2, 12, 3, 13), PlanData.fuelMin, self)
        fuelMin.step = 5
        fuelMin.leadZeros = 2
        self.entries.append(fuelMin)
        self.entries.append(LineEntry(self.rect(3, 12, 6, 13), \
                                      PlanData.alternate, self))
        self.entries.append(MultiEntry(self.rect(6, 11, 10, 13), \
                                       PlanData.pilot1, 3, self))
        self.entries.append(IntEntry(self.rect(10, 12, 11, 13), \
                                     PlanData.numAboard, self))
        self.entries.append(LineEntry(self.rect(0, 14, 4, 15), \
                                      PlanData.color, self))
        self.updatePlan()

    def rect(self, x1, y1, x2, y2):
        "Return QRect for given position"
        return QRect(QPoint(PlanView.colPos[x1], PlanView.rowPos[y1]), \
                     QPoint(PlanView.colPos[x2], PlanView.rowPos[y2]))

    def updatePlan(self):
        "Update plan contents and repaint"
        for entry in self.entries:
            entry.updateData()
        self.viewport().update()

    def updateMenus(self):
        "Signal main window to update menu status"
        self.emit(PYSIGNAL('menuChg'), ())

    def updateRoute(self):
        "Signal to update route for airspeed or depart time"
        self.emit(PYSIGNAL('paramChg'), ())

    def paintPlan(self, painter):
        "Paint flight plan"
        for coord in PlanView.vertLines:
            painter.drawLine(PlanView.colPos[coord[0]], \
                             PlanView.rowPos[coord[1]], \
                             PlanView.colPos[coord[0]], \
                             PlanView.rowPos[coord[2]])
        for coord in PlanView.horizLines:
            painter.drawLine(PlanView.colPos[coord[1]], \
                             PlanView.rowPos[coord[0]], \
                             PlanView.colPos[coord[2]], \
                             PlanView.rowPos[coord[0]])
        painter.setFont(self.smFont)
        painter.drawText(self.rect(1, 2, 2, 3), Qt.AlignCenter, 'VFR')
        painter.drawText(self.rect(1, 3, 2, 4), Qt.AlignCenter, 'IFR')
        painter.drawText(self.rect(1, 4, 2, 5), Qt.AlignCenter, 'DVFR')
        painter.save()
        painter.translate(0, 2)    # adjust title vertical positions
        painter.drawText(self.rect(0, 1, 2, 2), Qt.AlignHCenter, '1. TYPE')
        painter.drawText(self.rect(8, 1, 10, 2), Qt.AlignHCenter, \
                         '6. DEPARTURE TIME')
        painter.drawText(self.rect(8, 2, 9, 3), Qt.AlignHCenter, \
                         'PROPOSED (Z)')
        painter.drawText(self.rect(9, 2, 10, 3), Qt.AlignHCenter, \
                         'ACTUAL (Z)')
        painter.drawText(self.rect(4, 7, 6, 8), Qt.AlignHCenter, \
                         '10. EST. TIME ENROUTE')
        painter.drawText(self.rect(4, 8, 5, 9), Qt.AlignHCenter, 'HOURS')
        painter.drawText(self.rect(5, 8, 6, 9), Qt.AlignHCenter, 'MINUTES')
        painter.drawText(self.rect(0, 10, 3, 11), Qt.AlignHCenter, \
                         '12. FUEL ON BOARD')
        painter.drawText(self.rect(0, 11, 2, 12), Qt.AlignHCenter, 'HOURS')
        painter.drawText(self.rect(2, 11, 3, 12), Qt.AlignHCenter, 'MINUTES')
        painter.translate(2, 0)    # adjust title horizontal positions
        painter.drawText(self.rect(2, 1, 4, 3), 0, \
                         '2. AIRCRAFT\n   IDENTIFICATION')
        painter.drawText(self.rect(4, 1, 6, 3), 0, \
                         '3. AIRCRAFT TYPE/\n   SPECIAL EQUIPMENT')
        painter.drawText(self.rect(6, 1, 7, 3), 0, '4. TRUE\n   AIRSPEED')
        painter.drawText(self.rect(7, 1, 8, 3), 0, '5. DEPARTURE POINT')
        painter.drawText(self.rect(10, 1, 11, 3), 0, \
                         '7. CRUISING\n   ALTITUDE')
        painter.drawText(self.rect(0, 5, 11, 6), 0, '8. ROUTE OF FLIGHT')
        painter.drawText(self.rect(0, 7, 4, 9), 0, \
                         '9. DESTINATION (Name of airport\n   and city)')
        painter.drawText(self.rect(6, 7, 11, 9), 0, '11. REMARKS')
        painter.drawText(self.rect(3, 10, 6, 12), 0, \
                         '13. ALTERNATE AIRPORT(S)')
        painter.drawText(self.rect(6, 10, 10, 12), 0, \
              '14. PILOT\'S NAME, ADDRESS & PHONE & AIRCRAFT HOME BASE')
        painter.drawText(self.rect(10, 10, 11, 12), 0, '15. NUMBER\n   ABOARD')
        painter.drawText(self.rect(0, 13, 4, 14), 0, '16. COLOR OF AIRCRAFT')
        painter.restore()
        painter.setFont(self.normFont)
        painter.drawText(self.rect(0, 0, 11, 1), Qt.AlignCenter, 'FLIGHT PLAN')

        if FlyMainWin.route.wpList:    # depart & destination id
            painter.drawText(self.rect(7, 3, 8, 5), Qt.AlignCenter, \
                             FlyMainWin.route.wpList[0].ident())
            painter.drawText(self.rect(0, 9, 4, 10), Qt.AlignCenter, \
                             FlyMainWin.route.wpList[-1].ident())

        if FlyMainWin.route.legList:   # time enroute
            ete = FlyMainWin.route.legList[-1].totalTime.hourMinute()
            painter.drawText(self.rect(4, 9, 5, 10), Qt.AlignCenter, `ete[0]`)
            painter.drawText(self.rect(5, 9, 6, 10), Qt.AlignCenter, \
                             '%02d' % ete[1])
        for entry in self.entries:
            entry.drawText(painter)

    def drawContents(self, painter, clipX=None, clipY=None, \
                     clipW=None, clipH=None):
        "Paint view"
        if clipX == None:   # none is for wrong overridden function
            return QFrame.drawContents(self, painter)
        pal = self.palette()
        painter.setPen(pal.color(QPalette.Active, QColorGroup.Text))
        painter.save()
        painter.translate(PlanView.viewMargin, PlanView.viewMargin)
        self.paintPlan(painter)
        self.selectedEntry.drawSelected(painter)
        painter.restore()

    def contentsMousePressEvent(self, event):
        "Mouse entry selection control"
        position = QPoint(event.pos().x() - PlanView.viewMargin, \
                          event.pos().y() - PlanView.viewMargin)
        for entry in self.entries:
            if entry.rect.contains(position):
                if self.selectedEntry != entry:
                    self.selectedEntry = entry
                    self.viewport().update()
                return

    def contentsMouseDoubleClickEvent(self, event):
        "Mouse entry edit control"
        position = QPoint(event.pos().x() - PlanView.viewMargin, \
                          event.pos().y() - PlanView.viewMargin)
        for entry in self.entries:
            if entry.rect.contains(position):
                entry.editEntry()
                return

    def keyPressEvent(self, event):
        "Key bindings"
        if event.key() == Qt.Key_Up or event.key() == Qt.Key_Left:
            i = self.entries.index(self.selectedEntry) - 1
            if i < 0:
                i = len(self.entries) - 1
            self.selectedEntry = self.entries[i]
            self.viewport().update()
        elif event.key() == Qt.Key_Down or event.key() == Qt.Key_Right:
            i = self.entries.index(self.selectedEntry) + 1
            if i >= len(self.entries):
                i = 0
            self.selectedEntry = self.entries[i]
            self.viewport().update()
        elif event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
            self.selectedEntry.editEntry()
        else:
            QWidget.keyPressEvent(self, event)

    def closeEvent(self, event):
        "Signal that view is closing"
        self.emit(PYSIGNAL('viewClosed'), ())
        event.accept()


class PlanEntry(QObject):
    "Draw and edit entries in flight plan"
    def __init__(self, rect, dataTag=None, parent=None, name=None):
        QObject.__init__(self, parent, name)
        self.rect = rect
        self.dataTag = dataTag
        self.dataStr = ''
        self.ctrl = None

    def updateData(self):
        "Refresh data from PlanData"
        if self.dataTag != None:
            self.dataStr = FlyMainWin.route.fltPlan.data[self.dataTag]

    def saveData(self):
        "Store data to PlanData"
        if self.dataTag != None:
            FlyMainWin.route.fltPlan.data[self.dataTag] = self.dataStr

    def drawText(self, painter):
        "Draw text in rect"
        painter.drawText(self.rect, Qt.AlignCenter, self.dataStr)

    def drawSelected(self, painter):
        "Draw bold border for selected item"
        oldPen = painter.pen()
        painter.setPen(QPen(oldPen.color(), 3))
        painter.drawRect(self.rect)
        painter.setPen(oldPen)

    def editEntry(self):
        "Called for user edit of items"
        pass

    def ctrlRect(self):
        "Return adjusted rect for ctrl"
        rect = QRect(self.rect)
        rect.moveTopLeft(self.parent().contentsToViewport(rect.topLeft()))
        rect.moveBy(PlanView.viewMargin, PlanView.viewMargin)
        return rect


class BoolEntry(PlanEntry):
    "Draw and edit entries for types in flight plan"
    def __init__(self, rect, dataTag=None, boolNum=0, parent=None, name=None):
        PlanEntry.__init__(self, rect, dataTag, parent, name)
        self.boolNum = boolNum
        self.updateData()

    def editEntry(self):
        "Called for user edit of items"
        if not self.dataStr:
            self.dataStr = 'X'
            self.saveData()
            self.emit(PYSIGNAL('checked'), (self.boolNum,))
            self.parent().viewport().update()
            FlyMainWin.docModified = 1
            self.parent().updateMenus()

    def uncheck(self, boolNum):
        "Unset item in response to other change"
        if boolNum != self.boolNum:
            self.dataStr = ''
            self.saveData()
            self.emit(PYSIGNAL('checked'), (boolNum,))
            self.parent().viewport().update()


class LineEntry(PlanEntry):
    "Draw and edit entries for single text lines in flight plan"
    def __init__(self, rect, dataTag=None, parent=None, name=None):
        PlanEntry.__init__(self, rect, dataTag, parent, name)
        self.updateData()

    def editEntry(self):
        "Called for user edit of line items"
        self.ctrl = TmpEdit(self.dataStr, self.parent().viewport())
        self.ctrl.setGeometry(self.ctrlRect())
        self.ctrl.show()
        self.ctrl.setFocus()
        self.connect(self.ctrl, PYSIGNAL('editDone'), self.editDone)

    def editDone(self):
        "Called by ctrl after return or focus change"
        text = str(self.ctrl.text())
        self.ctrl.close(1)
        self.ctrl = None
        if text != self.dataStr:
            self.dataStr = text
            self.saveData()
            self.parent().viewport().update()
            FlyMainWin.docModified = 1
            self.parent().updateMenus()


class MultiEntry(PlanEntry):
    "Draw and edit entries for multiple text lines in flight plan"
    def __init__(self, rect, dataTag=None, numLines=4, parent=None, name=None):
        PlanEntry.__init__(self, rect, dataTag, parent, name)
        self.numLines = numLines
        self.updateData()

    def updateData(self):
        "Refresh  data from PlanData"
        if self.dataTag != None:
            self.dataStr = '\n'.join([FlyMainWin.route.fltPlan.data[num] \
                                      for num in range(self.dataTag, \
                                                       self.dataTag \
                                                       + self.numLines)])
            self.dataStr = self.dataStr.rstrip()

    def saveData(self):
        "Store data to PlanData"
        if self.dataTag != None:
            list = self.dataStr.split('\n', self.numLines)
            list.extend([''] * self.numLines)
            for num in range(self.numLines):
                FlyMainWin.route.fltPlan.data[num + self.dataTag] = list[num]

    def drawText(self, painter):
        "Draw text in rect"
        painter.save()
        painter.translate(2, 2)   # adjust position
        painter.drawText(self.rect, Qt.AlignTop | Qt.AlignLeft \
                         | Qt.WordBreak, self.dataStr)
        painter.restore()

    def editEntry(self):
        "Called for user edit of lines"
        self.ctrl = TmpMultiEdit(self.dataStr, self.parent().viewport())
        self.ctrl.setMaxLines(self.numLines)
        self.ctrl.setGeometry(self.ctrlRect())
        self.ctrl.show()
        self.ctrl.setFocus()
        self.connect(self.ctrl, PYSIGNAL('editDone'), self.editDone)

    def editDone(self):
        "Called by ctrl after return or focus change"
        text = str(self.ctrl.text()).rstrip()
        self.ctrl.close(1)
        self.ctrl = None
        if text != self.dataStr:
            self.dataStr = text
            self.saveData()
            self.parent().viewport().update()
            FlyMainWin.docModified = 1
            self.parent().updateMenus()


class IntEntry(PlanEntry):
    "Draw and edit entries for integers in flight plan"
    def __init__(self, rect, dataTag=None, parent=None, name=None):
        PlanEntry.__init__(self, rect, dataTag, parent, name)
        self.min = 0
        self.max = 99
        self.step = 1
        self.leadZeros = 0
        if self.dataTag != None:
            self.min = FlyMainWin.route.fltPlan.intMinList[self.dataTag]
            self.max = FlyMainWin.route.fltPlan.intMaxList[self.dataTag]
        self.updateData()

    def drawText(self, painter):
        "Draw int text in rect"
        if self.leadZeros == 2:
            painter.drawText(self.rect, Qt.AlignCenter, \
                             '%02d' % int(self.dataStr))
        else:
            painter.drawText(self.rect, Qt.AlignCenter, self.dataStr)

    def editEntry(self):
        "Called for user edit of int items"
        self.ctrl = SpinBoxEx(self.min, self.max, self.step, \
                              self.leadZeros, self.parent().viewport())
        self.ctrl.setValue(int(self.dataStr))
        self.ctrl.origValue = int(self.dataStr)
        self.ctrl.setGeometry(self.ctrlRect())
        self.ctrl.show()
        self.ctrl.setFocus()
        self.connect(self.ctrl, PYSIGNAL('editDone'), self.editDone)

    def editDone(self):
        "Called by ctrl after return or focus change"
        text = str(self.ctrl.text())
        self.ctrl.close(1)
        self.ctrl = None
        if text != self.dataStr:
            self.dataStr = text
            self.saveData()
            self.parent().viewport().update()
            FlyMainWin.docModified = 1
            self.parent().updateMenus()


class SpeedEntry(IntEntry):
    "Draw and edit the entry for airspeed in flight plan"
    def __init__(self, rect, dataTag=None, parent=None, name=None):
        IntEntry.__init__(self, rect, dataTag, parent, name)
        self.step = 5
        self.leadZeros = 0

    def updateData(self):
        "Refresh  data from route leg"
        self.min = 1
        self.max = RouteLeg.maxAirspeed
        self.dataStr = `FlyMainWin.option.intData('Airspeed', self.min, \
                                                  self.max)`

    def editDone(self):
        "Called by ctrl after return or focus change"
        text = str(self.ctrl.text())
        self.ctrl.close(1)
        self.ctrl = None
        if text != self.dataStr:
            self.dataStr = text
            self.parent().viewport().update()
            FlyMainWin.option.changeData('Airspeed', text, 1)
            FlyMainWin.option.writeChanges()
            self.parent().updateRoute()


class DepartEntry(PlanEntry):
    "Draw and edit entry for depart time in flight plan"
    def __init__(self, rect, dataTag=None, parent=None, name=None):
        PlanEntry.__init__(self, rect, dataTag, parent, name)
        self.updateData()

    def updateData(self):
        "Refresh  data from route leg"
        depart = Time(FlyMainWin.option.intData('DepartHour', 0, 23) \
                   + FlyMainWin.option.intData('DepartMinute', 0, 59) / 60.0)
        if FlyMainWin.option.boolData('LocalClock'):
            depart.toGMT()
        self.dataStr = depart.timeStr()

    def editEntry(self):
        "Called for user edit of departure"
        dlg = TimeDlg(self.parent().viewport(), None, 1)
        if dlg.exec_loop() == QDialog.Accepted:
            self.updateData()
            self.parent().viewport().update()
            self.parent().updateRoute()


class MapView(QWidget):
    "A view that shows a rough map of the route"
    def __init__(self, parent=None, name=None, flags=0):
        QWidget.__init__(self, parent, name, flags)
        self.setFocusPolicy(QWidget.StrongFocus)
        if sys.platform == 'win32':
            self.setCaption('Route Map  (PyQt)')
        else:
            self.setCaption('Route Map')
        self.setIcon(QPixmap(fly_xpm))
        topLayout = QVBoxLayout(self, 5)
        ctrlLayout = QHBoxLayout(topLayout)
        self.infoLabel = QLabel(self)
        self.infoLabel.setFrameStyle(QFrame.Panel | QFrame.Sunken)
        self.mapDrawing = MapDrawing(self.infoLabel, self)
        topLayout.addWidget(self.mapDrawing, 1)
        topLayout.addWidget(self.infoLabel)
        self.airportButton = QToolButton(self)
        self.airportButton.setTextLabel('Show All Airports')
        self.airportButton.setUsesTextLabel(1)
        self.airportButton.setFocusPolicy(QWidget.TabFocus)
        self.airportButton.setPixmap(QPixmap(map_apt_xpm))
        self.airportButton.setToggleButton(1)
        self.connect(self.airportButton, SIGNAL('toggled(bool)'), \
                     self.mapDrawing.addAirports)
        ctrlLayout.addWidget(self.airportButton)
        self.navaidButton = QToolButton(self)
        self.navaidButton.setTextLabel('Show All Navaids')
        self.navaidButton.setUsesTextLabel(1)
        self.navaidButton.setFocusPolicy(QWidget.TabFocus)
        self.navaidButton.setPixmap(QPixmap(map_nav_xpm))
        self.navaidButton.setToggleButton(1)
        self.connect(self.navaidButton, SIGNAL('toggled(bool)'), \
                     self.mapDrawing.addNavaids)
        ctrlLayout.addWidget(self.navaidButton)
        self.fixesButton = QToolButton(self)
        self.fixesButton.setTextLabel('Show All Fixes')
        self.fixesButton.setUsesTextLabel(1)
        self.fixesButton.setFocusPolicy(QWidget.TabFocus)
        self.fixesButton.setPixmap(QPixmap(map_fix_xpm))
        self.fixesButton.setToggleButton(1)
        self.connect(self.fixesButton, SIGNAL('toggled(bool)'), \
                     self.mapDrawing.addFixes)
        ctrlLayout.addWidget(self.fixesButton)
        self.mapDrawing.updateAll()

    def updateMap(self):
        "Update the map data and view"
        FlyMainWin.route.mapData.calcBounds()
        FlyMainWin.route.mapData.addRouteWp()
        if self.mapDrawing.inclAirports:
            FlyMainWin.route.mapData.addAirports()
        if self.mapDrawing.inclNavaids:
            FlyMainWin.route.mapData.addNavaids()
        if self.mapDrawing.inclFixes:
            FlyMainWin.route.mapData.addFixes()
        self.mapDrawing.updateAll()

    def closeEvent(self, event):
        "Signal that view is closing"
        self.emit(PYSIGNAL('viewClosed'), ())
        event.accept()

    def hideEvent(self, event):
        "Reset show all on map close"
        self.mapDrawing.inclAirports = 0
        self.mapDrawing.inclNavaids = 0
        self.mapDrawing.inclFixes = 0
        self.airportButton.setOn(0)
        self.navaidButton.setOn(0)
        self.fixesButton.setOn(0)


class MapDrawing(QWidget):
    "Shows a rough map of the route"
    mapBorder = 10
    labelOffset = 10
    infoDistance = 10
    def __init__(self, infoLabel, parent=None, name=None, flags=0):
        QWidget.__init__(self, parent, name, flags)
        self.infoLabel = infoLabel
        self.setMouseTracking(1)
        self.setMinimumSize(300, 300)
        self.inclAirports = 0
        self.inclNavaids = 0
        self.inclFixes = 0
        self.fullItemList = []

    def sizeHint(self):
        "Set prefered size"
        return QSize(MapData.mapPixels + 2 * MapDrawing.mapBorder, \
                     MapData.mapPixels + 2 * MapDrawing.mapBorder)

    def updateAll(self):
        "Update full item list and the map"
        self.fullItemList = FlyMainWin.route.mapData.routeList[:]
        if self.inclAirports:
            self.fullItemList.extend(FlyMainWin.route.mapData.airportList)
        if self.inclNavaids:
            self.fullItemList.extend(FlyMainWin.route.mapData.navaidList)
        if self.inclFixes:
            self.fullItemList.extend(FlyMainWin.route.mapData.fixList)
        self.update()

    def addAirports(self, show):
        "Add or remove other airports from map"
        self.inclAirports = show
        if show:
            FlyMainWin.route.mapData.addAirports()
        self.updateAll()

    def addNavaids(self, show):
        "Add or remove other navaids from map"
        self.inclNavaids = show
        if show:
            FlyMainWin.route.mapData.addNavaids()
        self.updateAll()

    def addFixes(self, show):
        "Add or remove other airports from map"
        self.inclFixes = show
        if show:
            FlyMainWin.route.mapData.addFixes()
        self.updateAll()

    def paintEvent(self, event):
        "Paint the map"
        paint = QPainter()   # problem in win with QPainter(self)
        paint.begin(self)
        paint.setWindow(0, 0, MapData.mapPixels + 2 * MapDrawing.mapBorder, \
                        MapData.mapPixels + 2 * MapDrawing.mapBorder)
        viewRect = paint.viewport()
        size = min(viewRect.width(), viewRect.height())
        paint.setViewport(viewRect.left() + (viewRect.width() - size) / 2, \
                          viewRect.top() + (viewRect.height() - size) / 2, \
                          size, size)
        paint.setPen(QPen(QColor(0, 0, 0), 2))
        paint.drawRect(paint.window())
        paint.translate(MapDrawing.mapBorder, MapDrawing.mapBorder)
        if FlyMainWin.route.mapData.routeList:
            if self.inclFixes:
                for item in FlyMainWin.route.mapData.fixList:
                    paint.drawPixmap(item.xCoord - 3, item.yCoord - 3, \
                                     QPixmap(map_fix_xpm))
            if self.inclNavaids:
                for item in FlyMainWin.route.mapData.navaidList:
                    paint.drawPixmap(item.xCoord - 3, item.yCoord - 3, \
                                     QPixmap(map_nav_xpm))
            if self.inclAirports:
                for item in FlyMainWin.route.mapData.airportList:
                    paint.drawPixmap(item.xCoord - 3, item.yCoord - 3, \
                                     QPixmap(map_apt_xpm))
            item = FlyMainWin.route.mapData.routeList[0]
            paint.drawPixmap(item.xCoord - 3, item.yCoord - 3, \
                             self.getPixmap(item))
            self.drawName(paint, item)
            paint.moveTo(item.xCoord, item.yCoord)
            for item in FlyMainWin.route.mapData.routeList[1:]:
                paint.lineTo(item.xCoord, item.yCoord)
                paint.drawPixmap(item.xCoord - 3, item.yCoord - 3, \
                                 self.getPixmap(item))
                self.drawName(paint, item)
        paint.end()

    def getPixmap(self, item):
        "Return pixmap for a given map item"
        if item.waypoint.__class__ == Airport:
            return QPixmap(map_apt_xpm)
        elif item.waypoint.__class__ == Navaid:
            return QPixmap(map_nav_xpm)
        else:
            return QPixmap(map_fix_xpm)

    def drawName(self, paint, item):
        "Draw ID as legend for item"
        textRect = QFontMetrics(self.font()).boundingRect(item.\
                                                          waypoint.ident())
        pos = [item.xCoord, item.yCoord + textRect.height() / 2]
        if item.labelOctant in (0, 4):
            pos[0] -= textRect.width() / 2
        elif item.labelOctant in (5, 7):
            pos[0] -= textRect.width()
        elif item.labelOctant == 2:
            pos[0] += MapDrawing.labelOffset
        elif item.labelOctant == 6:
            pos[0] -= textRect.width() + MapDrawing.labelOffset
        if item.labelOctant in (0, 1, 7):
            pos[1] -= MapDrawing.labelOffset
        elif item.labelOctant in (3, 4, 5):
            pos[1] += MapDrawing.labelOffset
        paint.drawText(pos[0], pos[1], item.waypoint.ident())

    def calcSqDist(self, point, item):
        "Return distance in pixels from QPoint to item"
        return (point.x() - item.xCoord)**2 + (point.y() - item.yCoord)**2

    def mouseMoveEvent(self, event):
        "Called to track mouse & update waypoint label"
        pos = QPoint(event.x() - MapDrawing.mapBorder, \
                     event.y() - MapDrawing.mapBorder)
        distList = [self.calcSqDist(pos, item) for item in self.fullItemList]
        text = ''
        if distList and min(distList) <= MapDrawing.infoDistance**2:
            wp = self.fullItemList[distList.index(min(distList))].waypoint
            text = '%s - %s' % (wp.ident(), wp.description())
        self.infoLabel.setText(text)


class FlyMainWin(QMainWindow):
    "Main window, menus, toolbar, and status"
    option = Option('flyway', 20)
    route = Route(option)
    routePopup = None
    docModified = 0
    toolBar = None
    minPrintMargin = 0.25
    maxPrintMargin = 3.0
    printFontHt = 14
    printGap = 44
    defaultList = [\
        "# Options for FlyWay, a flight planning program",
        "#",
        "# All options can be set from within the program,",
        "# editing here is not recommended",
        "#",
        "Airspeed            100",
        "12hrClock           yes",
        "LocalClock          yes",
        "#",
        "PrintPrev           yes",
        "HorizMargin         0.5",
        "VertMargin          0.7",
        "PrintRoute          yes",
        "PrintFltPlan        yes",
        "PrintWaypoints      yes",
        "#",
        "AddOnEnter          yes",
        "DelKey              yes",
        "PlanShowOnOpen      yes",
        "#",
        "BackgroundR         255",
        "BackgroundG         255",
        "BackgroundB         255",
        "ForegroundR         0",
        "ForegroundG         0",
        "ForegroundB         0",
        "#",
        "RecentFile1         ",
        "RecentFile2         ",
        "RecentFile3         ",
        "RecentFile4         ",
        "#",
        "PlanType1           ",
        "PlanType2           X",
        "PlanType3           ",
        "PlanCraftID         ",
        "PlanEquip           ",
        "PlanAltitude        3500",
        "PlanRemark1         ",
        "PlanRemark2         ",
        "PlanRemark3         ",
        "PlanFuelHr          5",
        "PlanFuelMin         0",
        "PlanPilot1          ",
        "PlanPilot2          ",
        "PlanPilot3          ",
        "PlanNumAboard       1",
        "PlanColor           W"]
    def __init__(self, parent=None, name=None):
        QMainWindow.__init__(self, parent, name)
        FlyMainWin.option.loadAll(FlyMainWin.defaultList)
        FlyMainWin.option.addData('WindSpeed', '0')
        FlyMainWin.option.addData('WindDir', '360')
        FlyMainWin.option.addData('DepartHour', '8')
        FlyMainWin.option.addData('DepartMinute', '0')

        self.planView = None
        self.mapView = None
        self.helpView = None
        self.printer = QPrinter()
        self.printer.setPageSize(QPrinter.Letter)
        self.rtPageList = []
        self.planPageList = []
        self.wpPageList = []

        self.resize(540, 530)
        split = QSplitter(QSplitter.Vertical, self)
        self.setCentralWidget(split)
        self.rtView = RouteView(split)
        self.rtView.setFocus()
        self.wpView = WpView(split)
        split.setSizes([350, 180])
        self.connect(self.rtView, PYSIGNAL('selChange'), self.wpView.setSelWp)
        self.connect(self.rtView, PYSIGNAL('selChange'), self.setMenuAvail)
        self.connect(self.rtView, PYSIGNAL('recalcDone'), self.updatePlan)
        self.connect(self.rtView, PYSIGNAL('statusBarMsg'), self.statusBar(), \
                     SLOT('message(QString, int)'))

        if sys.platform == 'win32':
            self.setCaption('FlyWay  (PyQt)')
        else:
            self.setCaption('FlyWay')
        self.setIcon(QPixmap(fly_xpm))
        self.updateColors()
        self.fileName = ''
        self.statusBar().message('Hit return to add a waypoint...', 100000)

        self.setupMenus()
        self.setMenuAvail()
        self.fltPlnGrp.setEnabled(0)

    def updatePlan(self):
        "Update flight plan, map & menus"
        if self.planView and self.planView.isVisible():
            self.planView.updatePlan()
        if self.mapView and self.mapView.isVisible():
            self.mapView.updateMap()
        self.setMenuAvail()

    def setMenuAvail(self):
        "Update menu availability"
        self.fileSaveAct.setEnabled(FlyMainWin.docModified)
        self.editWpGrp.setEnabled(len(FlyMainWin.route.wpList))
        self.editUpAct.setEnabled(self.rtView.numSelect)
        self.editDownAct.setEnabled(self.rtView.numSelect < \
                                    len(FlyMainWin.route.wpList) - 1)
        self.editMultWpGrp.setEnabled(len(FlyMainWin.route.wpList) > 1)

    def updateColors(self):
        "Adjust the colors to the current option settings"
        pal = QApplication.palette()
        background = QColor(self.option.intData('BackgroundR', 0, 255), \
                            self.option.intData('BackgroundG', 0, 255), \
                            self.option.intData('BackgroundB', 0, 255))
        foreground = QColor(self.option.intData('ForegroundR', 0, 255), \
                            self.option.intData('ForegroundG', 0, 255), \
                            self.option.intData('ForegroundB', 0, 255))
        pal.setColor(QColorGroup.Base, background)
        pal.setColor(QPalette.Active, QColorGroup.Text, foreground)
        pal.setColor(QPalette.Inactive, QColorGroup.Text, foreground)
        QApplication.setPalette(pal, 1)
        pal.setColor(QColorGroup.Text, QColor(0, 0, 0))
        self.statusBar().setPalette(pal)

    def openFile(self, fileName):
        "Open given file"
        if FlyMainWin.route.readFile(fileName):
            if FlyMainWin.route.wpErrorList:
                QMessageBox.warning(self, 'FlyWay', \
                                    'Error - could not find the '\
                                    'following waypoints:\n%s' % '\n'.\
                                    join(FlyMainWin.route.wpErrorList))
            if sys.platform == 'win32':
                self.setCaption(fileName + ' - FlyWay  (PyQt)')
            else:
                self.setCaption(fileName + ' - FlyWay')
            self.fileName = fileName
            self.recentFiles.addEntry(fileName)
            FlyMainWin.docModified = 0
            self.rtView.numSelect = 0
            self.rtView.recalc()
            if FlyMainWin.route.fltPlan \
               and FlyMainWin.option.boolData('PlanShowOnOpen'):
                self.fltPlnShow(1)
            self.statusBar().message('Ready...', 4000)
            return 1
        return 0

    def recentOpen(self, filePath):
        "Open from recentFiles signal"
        if filePath and self.savePrompt():
            if not self.openFile(filePath):
                self.recentFiles.removeEntry(filePath)

    def fileNew(self):
        "New file command"
        if self.savePrompt():
            FlyMainWin.route = Route(FlyMainWin.option)
            if sys.platform == 'win32':
                self.setCaption('FlyWay  (PyQt)')
            else:
                self.setCaption('FlyWay')
            self.fileName = ''
            FlyMainWin.docModified = 0
            self.rtView.numSelect = 0
            self.rtView.recalc()
            self.statusBar().message('Hit return to add a waypoint...', 100000)

    def fileOpen(self):
        "Open a file"
        self.statusBar().message('', 1)
        if self.savePrompt():
            fileName = str(QFileDialog.getOpenFileName(None, \
                           'FlyWay Files (*.fly);;All Files (*)', self))
            if fileName:
                self.openFile(fileName)

    def fileSave(self):
        "Save current file"
        if self.fileName:
            FlyMainWin.route.writeFile(self.fileName)
            FlyMainWin.docModified = 0
            self.setMenuAvail()
        else:
            self.fileSaveAs()

    def fileSaveAs(self):
        "Save file with a new name"
        dir, name = os.path.split(self.fileName)
        dlg = QFileDialog(dir, 'FlyWay Files (*.fly);;All Files (*)', \
                          self, '', 1)
        dlg.setMode(QFileDialog.AnyFile)
        dlg.setSelection(name)
        if sys.platform == 'win32':
            dlg.setCaption('Save As  (PyQt)')
        else:
            dlg.setCaption('Save As')
        dlg.setIcon(QPixmap(fly_xpm))
        result = dlg.exec_loop()
        fileName = str(dlg.selectedFile())
        if result == QDialog.Accepted and fileName:
            if '.' not in fileName and \
               str(dlg.selectedFilter()) == 'FlyWay Files (*.fly)':
                fileName += '.fly'
            if os.access(fileName, os.F_OK):
                text = 'File ' + fileName + ' exists\nReplace existing file?'
                ans = QMessageBox.information(self, 'FlyWay', text, \
                                              'Yes', 'No', 'Cancel', 0, 2)
                if ans != 0:
                    if ans == 1:
                        self.fileSaveAs()
                    return
            if FlyMainWin.route.writeFile(fileName):
                if sys.platform == 'win32':
                    self.setCaption(fileName + ' - FlyWay  (PyQt)')
                else:
                    self.setCaption(fileName + ' - FlyWay')
                self.fileName = fileName
                self.recentFiles.addEntry(fileName)
                FlyMainWin.docModified = 0
                self.setMenuAvail()

    def printPage(self, pageNum, painter):
        "Send pageNum to painter"
        if FlyMainWin.option.boolData('PrintRoute') \
           and len(self.rtPageList) > pageNum - 1:
            endLine = None
            if len(self.rtPageList) > pageNum:
                endLine = self.rtPageList[pageNum][0]
            painter.save()
            painter.translate(-RouteView.margin, \
                              self.rtPageList[pageNum - 1][1])
            self.rtView.paintTable(painter, self.rtPageList[pageNum - 1][0], \
                                   endLine)
            painter.restore()
        if self.planView and FlyMainWin.option.boolData('PrintFltPlan') \
           and len(self.planPageList) > pageNum - 1 \
           and self.planPageList[pageNum - 1]:
            painter.save()
            painter.translate(0, self.planPageList[pageNum - 1])
            self.planView.paintPlan(painter)
            painter.restore()
        if FlyMainWin.option.boolData('PrintWaypoints') \
           and self.wpPageList[pageNum - 1]:
            for wpGroup in self.wpPageList[pageNum - 1]:
                painter.save()
                painter.translate(-WpView.idStart[0], wpGroup[1] \
                                  - WpView.idStart[1])
                self.wpView.paintWp(painter, wpGroup[0])
                painter.restore()

    def filePrint(self):
        "Print file"
        useFile = self.printer.outputToFile()
        self.printer.setOutputFileName(os.path.splitext(self.fileName)[0] + \
                                       '.ps')
        self.printer.setOutputToFile(useFile)
        self.printer.setFullPage(1)
        self.printer.setMinMax(1, 99)
        self.printer.setFromTo(1, 99)
        if not self.printer.setup(self):
            return
        paintMetrics = QPaintDeviceMetrics(self.printer)
        dpi = min(paintMetrics.logicalDpiX(), paintMetrics.logicalDpiY())
        scale = dpi / 72.0
        xMargin = int(max(self.printer.margins().width() / scale, \
                          72 * self.option.numData('HorizMargin', \
                                                   FlyMainWin.minPrintMargin, \
                                                   FlyMainWin.maxPrintMargin)))
        yMargin = int(max(self.printer.margins().height() / scale, \
                          72 * self.option.numData('VertMargin', \
                                                   FlyMainWin.minPrintMargin, \
                                                   FlyMainWin.maxPrintMargin)))
        pageSize = (int(paintMetrics.width() / scale), \
                    int(paintMetrics.height() / scale))
        pageHeight = pageSize[1] - 2 * yMargin
        pageUsed = 0
        self.rtPageList = []
        if FlyMainWin.option.boolData('PrintRoute'):
            self.rtPageList = [(-1, 0)]     # line num, y-pos offset
            pageUsed += RouteView.wpStart[1]
            for lineNum in range(len(FlyMainWin.route.wpList) * 2 - 1):
                pageUsed += RouteView.rowSize
                if pageUsed > pageHeight:            # new page
                    self.rtPageList.append((lineNum, FlyMainWin.printFontHt \
                                            - lineNum * RouteView.rowSize \
                                            - RouteView.wpStart[1]))
                    pageUsed = FlyMainWin.printFontHt
        self.planPageList = []
        if self.planView and FlyMainWin.option.boolData('PrintFltPlan'):
            self.planPageList = [None] * (len(self.rtPageList) - 1)
            pos = pageUsed + FlyMainWin.printGap
            pageUsed = pos + PlanView.rowPos[-1]
            if pageUsed > pageHeight:            # new page
                self.planPageList.append(None)
                pos = FlyMainWin.printFontHt
                pageUsed = pos + PlanView.rowPos[-1]
            self.planPageList.append(pos)
        if FlyMainWin.option.boolData('PrintWaypoints'):
            self.wpPageList = [[]] * (max(len(self.rtPageList), \
                                          len(self.planPageList)) - 1)
            thisPage = []
            for wp in FlyMainWin.route.wpList:
                pos = pageUsed + FlyMainWin.printGap
                pageUsed = pos + (wp.detailNumLines() - 1) * WpView.rowSize
                if pageUsed > pageHeight:            # new page
                    self.wpPageList.append(thisPage)
                    thisPage = []
                    pos = FlyMainWin.printFontHt
                    pageUsed = pos + (wp.detailNumLines() - 1) * WpView.rowSize
                thisPage.append((wp, pos))
            self.wpPageList.append(thisPage)
        minPg = max(self.printer.fromPage(), 1)
        maxPg = max(len(self.rtPageList), len(self.wpPageList))
        if 1 <= self.printer.toPage() < maxPg:
            maxPg = self.printer.toPage()

        if self.option.boolData('PrintPrev'):
            dlg = PrintPrevDlg(minPg, maxPg, pageSize, (xMargin, yMargin),
                               self.printPage, self, None, 1)
            dlg.setIcon(QPixmap(fly_xpm))
            if dlg.exec_loop() != QDialog.Accepted:
                return
        self.printer.setFullPage(0)  # try to fix MS margin problem
        painter = QPainter()
        if painter.begin(self.printer):
            if self.printer.pageOrder() == QPrinter.FirstPageFirst:
                seq = range(minPg, maxPg + 1)
            else:
                seq = range(maxPg, minPg - 1, -1)
            printMargin = ((pageSize[0] * scale - paintMetrics.width()) / 2, \
                           (pageSize[1] * scale - paintMetrics.height()) / 2)
            pageAvail = (pageSize[0] - 2 * xMargin, pageHeight)
            painter.setWindow(0, 0, pageAvail[0], pageAvail[1])
            painter.setViewport(xMargin * scale -printMargin[0], \
                                yMargin * scale -printMargin[1], \
                                pageAvail[0] * scale, pageAvail[1] * scale)
            self.printPage(seq.pop(0), painter)
            for num in seq:
                self.printer.newPage()
                self.printPage(num, painter)
            painter.end()

    def filePrintOpt(self):
        "Set margins and page size for printing"
        dlg = OptionDlg(self.option, self, None, 1)
        if sys.platform == 'win32':
            dlg.setCaption('Print Options  (PyQt)')
        else:
            dlg.setCaption('Print Options')
        dlg.setIcon(QPixmap(fly_xpm))

        dlg.startGroupBox('Print Preview')
        OptionDlgBool(dlg, 'PrintPrev', 'Preview before printing')
        dlg.startGroupBox('Margins')
        OptionDlgDbl(dlg, 'HorizMargin', 'Horizontal page margins (inch)', \
                     FlyMainWin.minPrintMargin, FlyMainWin.maxPrintMargin)
        OptionDlgDbl(dlg, 'VertMargin', 'Vertical page margins (inch)', \
                     FlyMainWin.minPrintMargin, FlyMainWin.maxPrintMargin)
        dlg.startGroupBox('Include in Printout')
        OptionDlgBool(dlg, 'PrintRoute', 'Route Table')
        OptionDlgBool(dlg, 'PrintFltPlan', 'Flight Plan')
        OptionDlgBool(dlg, 'PrintWaypoints', 'Waypoint Info')

        if dlg.exec_loop() == QDialog.Accepted and self.option.chgList:
            self.option.writeChanges()

    def editParam(self):
        "Change airspeed and wind condition"
        dlg = OptionDlg(self.option, self, None, 1)
        if sys.platform == 'win32':
            dlg.setCaption('Parameters  (PyQt)')
        else:
            dlg.setCaption('Parameters')
        dlg.setIcon(QPixmap(fly_xpm))
        OptionDlgInt(dlg, 'Airspeed', 'Airspeed', 1, RouteLeg.maxAirspeed, \
                     1, 5, 0, 0, ' kts')
        OptionDlgInt(dlg, 'WindSpeed', 'Wind Speed', 0, RouteLeg.maxWindSpeed,\
                     0, 1, 0, 0, ' kts')
        OptionDlgInt(dlg, 'WindDir', 'Wind Direction', 0, 360, \
                     0, 5, 1, 3, chr(176))
        if dlg.exec_loop() == QDialog.Accepted:
            FlyMainWin.option.writeChanges()
            self.rtView.recalc()

    def editDepart(self):
        "Change departure time"
        dlg = TimeDlg(self, None, 1)
        if dlg.exec_loop() == QDialog.Accepted:
            self.rtView.recalc()

    def fltPlnShow(self, show):
        "Toggle the flight plan window"
        if show:
            if not self.planView:
                if not FlyMainWin.route.fltPlan:
                    FlyMainWin.route.fltPlan = PlanData(self.option)
                    FlyMainWin.route.fltPlan.setToDefaults()
                    FlyMainWin.route.fltPlan.setDefaultRoute(FlyMainWin.route)
                self.planView = PlanView()
                self.connect(self.planView, PYSIGNAL('menuChg'), \
                             self.setMenuAvail)
                self.connect(self.planView, PYSIGNAL('viewClosed'), \
                             self.planMenuChg)
                self.connect(self.planView, PYSIGNAL('paramChg'), \
                             self.rtView.recalc)
            self.planView.show()
            self.planView.updatePlan()
            self.fltPlnGrp.setEnabled(1)
        else:
            self.planView.hide()
            self.fltPlnGrp.setEnabled(0)

    def planMenuChg(self):
        "Reset flight plan menu on plan view close"
        self.fltPlnShowAct.setOn(0)
        self.fltPlnGrp.setEnabled(0)

    def fltPlnUpdate(self):
        "Update defaults with current settings"
        FlyMainWin.route.fltPlan.updateDefaults()
        self.statusBar().message('Default settings have been updated', 4000)

    def fltPlnReset(self):
        "Reset flight plan to default settings"
        FlyMainWin.route.fltPlan.setToDefaults()
        FlyMainWin.route.fltPlan.setDefaultRoute(FlyMainWin.route)
        self.planView.updatePlan()

    def toolsDist(self):
        "Compare route to direct distance"
        if len(FlyMainWin.route.wpList) > 1:
            text =  'Current route distance:  %3.0f\n\n'
            text += 'Direct route distance:    %3.0f'
            text = text % (FlyMainWin.route.legList[-1].totalDist, \
                           RouteLeg(FlyMainWin.route.wpList[0], \
                                    FlyMainWin.route.wpList[-1], \
                                    FlyMainWin.option).totalDist)
            dlg = QMessageBox('Distance Comparison', text, \
                              QMessageBox.NoIcon, QMessageBox.Ok \
                              | QMessageBox.Default, 0, 0, self)
            dlg.setIcon(QPixmap(fly_xpm))
            dlg.exec_loop()

    def toolsMap(self, show):
        "Activate the map view"
        if show:
            if not self.mapView:
                if not FlyMainWin.route.mapData:
                    FlyMainWin.route.mapData = MapData(FlyMainWin.route)
                self.mapView = MapView()
                self.connect(self.mapView, PYSIGNAL('viewClosed'), \
                             self.mapMenuChg)
            else:
                self.mapView.updateMap()
            self.mapView.show()
        else:
            self.mapView.hide()

    def mapMenuChg(self):
        "Reset map menu item on map view close"
        self.toolsMapAct.setOn(0)

    def toolsPref(self):
        "Set user preferences"
        dlg = OptionDlg(FlyMainWin.option, self, None, 1)
        dlg.setIcon(QPixmap(fly_xpm))
        OptionDlgBool(dlg, 'AddOnEnter', 'Enter key inserts waypoint')
        OptionDlgBool(dlg, 'DelKey', 'Delete key removes waypoint')
        OptionDlgBool(dlg, 'PlanShowOnOpen', 'Show flight plan when opening')
        if dlg.exec_loop() == QDialog.Accepted and self.option.chgList:
            FlyMainWin.option.writeChanges()

    def toolsBkColor(self):
        "Set view background color"
        background = QColor(FlyMainWin.option.intData('BackgroundR', 0, 255), \
                            FlyMainWin.option.intData('BackgroundG', 0, 255), \
                            FlyMainWin.option.intData('BackgroundB', 0, 255))
        newColor = QColorDialog.getColor(background, self)
        if newColor.isValid() and newColor != background:
            FlyMainWin.option.changeData('BackgroundR', `newColor.red()`, 1)
            FlyMainWin.option.changeData('BackgroundG', `newColor.green()`, 1)
            FlyMainWin.option.changeData('BackgroundB', `newColor.blue()`, 1)
            FlyMainWin.option.writeChanges()
            self.updateColors()

    def toolsTxtColor(self):
        "Set view text color"
        foreground = QColor(FlyMainWin.option.intData('ForegroundR', 0, 255), \
                            FlyMainWin.option.intData('ForegroundG', 0, 255), \
                            FlyMainWin.option.intData('ForegroundB', 0, 255))
        newColor = QColorDialog.getColor(foreground, self)
        if newColor.isValid() and newColor != foreground:
            FlyMainWin.option.changeData('ForegroundR', `newColor.red()`, 1)
            FlyMainWin.option.changeData('ForegroundG', `newColor.green()`, 1)
            FlyMainWin.option.changeData('ForegroundB', `newColor.blue()`, 1)
            FlyMainWin.option.writeChanges()
            self.updateColors()

    def helpReadMe(self):
        "View the ReadMe file"
        if not self.helpView:
            try:
                f = open(os.path.join(os.path.abspath(sys.path[0]), \
                                      'README.html'), 'r')
            except IOError:
                self.statusBar().message('Read Me file not found', 4000)
                return
            self.helpView = HelpView('\n'.join(f.readlines()), \
                                     'FlyWay README File')
            self.helpView.setIcon(QPixmap(fly_xpm))
        self.helpView.show()

    def helpAbout(self):
        "About this program"
        QMessageBox.about(self, 'FlyWay', 'FlyWay Version 0.2.2\nby Doug Bell')

    def savePrompt(self):
        "Ask for save if doc modified, return false on cancel"
        if FlyMainWin.docModified:
            ans = QMessageBox.information(self, 'FlyWay', 'Save changes?', \
                                          'Yes', 'No', 'Cancel', 0, 2)
            if ans == 0:
                self.fileSave()
            elif ans == 2:
                return 0
        return 1

    def closeEvent(self, event):
        "Check if doc modified"
        if self.savePrompt():
            event.accept()
        else:
            event.ignore()

    def setupMenus(self):
        "Add menu and toolbar entries"
        fileMenu = QPopupMenu(self)
        self.menuBar().insertItem('&File', fileMenu)
        FlyMainWin.toolBar = QToolBar(self)
        FlyMainWin.routePopup = QPopupMenu(self.rtView)

        fileNewAct = QAction('New File', QIconSet(QPixmap(filenew_xpm)), \
                             '&New', Qt.CTRL+Qt.Key_N, self)
        fileNewAct.setStatusTip('Start a new file')
        fileNewAct.addTo(fileMenu)
        fileNewAct.addTo(FlyMainWin.toolBar)
        self.connect(fileNewAct, SIGNAL('activated()'), self.fileNew)

        fileOpenAct = QAction('Open File', QIconSet(QPixmap(fileopen_xpm)), \
                              '&Open...', Qt.CTRL+Qt.Key_O, self)
        fileOpenAct.setStatusTip('Open a file from disk')
        fileOpenAct.addTo(fileMenu)
        fileOpenAct.addTo(FlyMainWin.toolBar)
        self.connect(fileOpenAct, SIGNAL('activated()'), self.fileOpen)

        fileMenu.insertSeparator()

        self.fileSaveAct = QAction('Save File', \
                                   QIconSet(QPixmap(filesave_xpm)), \
                                   '&Save', Qt.CTRL+Qt.Key_S, self)
        self.fileSaveAct.setStatusTip('Save changes to the current file')
        self.fileSaveAct.addTo(fileMenu)
        self.fileSaveAct.addTo(FlyMainWin.toolBar)
        self.connect(self.fileSaveAct, SIGNAL('activated()'), self.fileSave)

        fileSaveAsAct = QAction('Save As', 'Save &As...', 0, self)
        fileSaveAsAct.setStatusTip('Save the file with a new name')
        fileSaveAsAct.addTo(fileMenu)
        self.connect(fileSaveAsAct, SIGNAL('activated()'), self.fileSaveAs)

        fileMenu.insertSeparator()
        FlyMainWin.toolBar.addSeparator()

        filePrintAct = QAction('Print File', \
                               QIconSet(QPixmap(fileprint_xpm)), \
                               '&Print...', Qt.CTRL+Qt.Key_P, self)
        filePrintAct.setStatusTip('Print the route info')
        filePrintAct.addTo(fileMenu)
        filePrintAct.addTo(FlyMainWin.toolBar)
        self.connect(filePrintAct, SIGNAL('activated()'), self.filePrint)

        filePrintOptAct = QAction('Print Options', 'P&rint Options...', \
                                  0, self)
        filePrintOptAct.setStatusTip('Set margins and page size for printing')
        filePrintOptAct.addTo(fileMenu)
        self.connect(filePrintOptAct, SIGNAL('activated()'), self.filePrintOpt)

        fileMenu.insertSeparator()
        self.recentFiles = RecentFiles(self.option, 4, 30, self)
        self.connect(self.recentFiles, PYSIGNAL('askOpen'), self.recentOpen)
        self.recentFiles.initMenu(fileMenu, 100)

        fileMenu.insertSeparator()
        FlyMainWin.toolBar.addSeparator()

        fileQuitAct = QAction('Quit', '&Quit', Qt.CTRL+Qt.Key_Q, self)
        fileQuitAct.setStatusTip('Exit the application')
        fileQuitAct.addTo(fileMenu)
        self.connect(fileQuitAct, SIGNAL('activated()'), self.close)

        editMenu = QPopupMenu(self)
        self.menuBar().insertItem('&Edit', editMenu)
        self.editWpGrp = QActionGroup(self)
        self.editMultWpGrp = QActionGroup(self)

        editInBeforeAct = QAction('Insert Waypoint Before', \
                                  'Insert Waypoint &Before', 0, self.editWpGrp)
        editInBeforeAct.setStatusTip('Add a new waypoint before the selection')
        editInBeforeAct.addTo(editMenu)
        editInBeforeAct.addTo(FlyMainWin.routePopup)
        rt = FlyMainWin.route
        self.connect(editInBeforeAct, SIGNAL('activated()'), \
                     self.rtView.editInBefore)

        editInAfterAct = QAction('Insert Waypoint After', \
                                 'Insert Waypoint &After\tRTN', 0, self)
        editInAfterAct.setStatusTip('Add a new waypoint after the selection')
        editInAfterAct.addTo(editMenu)
        editInAfterAct.addTo(FlyMainWin.routePopup)
        self.connect(editInAfterAct, SIGNAL('activated()'), \
                     self.rtView.editInAfter)

        editDeleteAct = QAction('Delete Waypoint', '&Delete Waypoint\tDEL', \
                                0, self.editWpGrp)
        editDeleteAct.setStatusTip('Delete the selected waypoint')
        editDeleteAct.addTo(editMenu)
        editDeleteAct.addTo(FlyMainWin.routePopup)
        self.connect(editDeleteAct, SIGNAL('activated()'), \
                     self.rtView.editDelete)

        editMenu.insertSeparator()
        FlyMainWin.routePopup.insertSeparator()

        self.editUpAct = QAction('Move Waypoint Up', 'Move Waypoint &Up', \
                                 0, self)
        self.editUpAct.setStatusTip('Move the selected waypoint upward')
        self.editUpAct.addTo(editMenu)
        self.editUpAct.addTo(FlyMainWin.routePopup)
        self.connect(self.editUpAct, SIGNAL('activated()'), \
                     self.rtView.editUp)

        self.editDownAct = QAction('Move Waypoint Down', \
                                   '&Move Waypoint Down', 0, self)
        self.editDownAct.setStatusTip('Move the selected waypoint downward')
        self.editDownAct.addTo(editMenu)
        self.editDownAct.addTo(FlyMainWin.routePopup)
        self.connect(self.editDownAct, SIGNAL('activated()'), \
                     self.rtView.editDown)

        self.editReverseAct = QAction('Reverse Route', '&Reverse Route', \
                                      0, self.editMultWpGrp)
        self.editReverseAct.setStatusTip('Reverse the order of all waypoints')
        self.editReverseAct.addTo(editMenu)
        self.connect(self.editReverseAct, SIGNAL('activated()'), \
                     self.rtView.editReverse)

        editMenu.insertSeparator()

        editParamAct = QAction('Change Parameters', \
                               QIconSet(QPixmap(editparam_xpm)), \
                               '&Change Parameters...', 0, self)
        editParamAct.setStatusTip('Change airspeed and wind condition')
        editParamAct.addTo(editMenu)
        editParamAct.addTo(FlyMainWin.toolBar)
        self.connect(editParamAct, SIGNAL('activated()'), self.editParam)

        editDepartAct = QAction('Change Departure Time', \
                                QIconSet(QPixmap(editdepart_xpm)), \
                               'Change Departure &Time...', 0, self)
        editDepartAct.setStatusTip('Change departure time')
        editDepartAct.addTo(editMenu)
        editDepartAct.addTo(FlyMainWin.toolBar)
        self.connect(editDepartAct, SIGNAL('activated()'), self.editDepart)

        fltPlnMenu = QPopupMenu(self)
        self.menuBar().insertItem('Flight &Plan', fltPlnMenu)
        FlyMainWin.toolBar.addSeparator()

        self.fltPlnShowAct = QAction('Show Flight Plan', \
                                     QIconSet(QPixmap(fltplanshow_xpm)), \
                                     '&Show Flight Plan', 0, self)
        self.fltPlnShowAct.setStatusTip('Toggle the flight plan window')
        self.fltPlnShowAct.setToggleAction(1)
        self.fltPlnShowAct.addTo(fltPlnMenu)
        self.fltPlnShowAct.addTo(FlyMainWin.toolBar)
        self.connect(self.fltPlnShowAct, SIGNAL('toggled(bool)'), \
                     self.fltPlnShow)

        fltPlnMenu.insertSeparator()
        self.fltPlnGrp = QActionGroup(self)

        fltPlnUpdateAct = QAction('Update Defaults', '&Update Defaults', \
                                  0, self.fltPlnGrp)
        fltPlnUpdateAct.setStatusTip('Update defaults with current settings')
        fltPlnUpdateAct.addTo(fltPlnMenu)
        self.connect(fltPlnUpdateAct, SIGNAL('activated()'), self.fltPlnUpdate)

        fltPlnResetAct = QAction('Reset to Defaults', '&Reset to Defaults', \
                                  0, self.fltPlnGrp)
        fltPlnResetAct.setStatusTip('Reset flight plan to default settings')
        fltPlnResetAct.addTo(fltPlnMenu)
        self.connect(fltPlnResetAct, SIGNAL('activated()'), self.fltPlnReset)

        toolsMenu = QPopupMenu(self)
        self.menuBar().insertItem('&Tools', toolsMenu)

        toolsDistAct = QAction('Distance Comparison', '&Distance Comparison', \
                               0, self.editMultWpGrp)
        toolsDistAct.setStatusTip('Compare route to direct distance')
        toolsDistAct.addTo(toolsMenu)
        self.connect(toolsDistAct, SIGNAL('activated()'), self.toolsDist)

        self.toolsMapAct = QAction('View Map', \
                                   QIconSet(QPixmap(mapshow_xpm)), \
                                   'View &Map', 0, self)
        self.toolsMapAct.setStatusTip('View a rough map of the route')
        self.toolsMapAct.setToggleAction(1)
        self.toolsMapAct.addTo(toolsMenu)
        self.toolsMapAct.addTo(FlyMainWin.toolBar)
        self.connect(self.toolsMapAct, SIGNAL('toggled(bool)'), self.toolsMap)

        toolsMenu.insertSeparator()

        toolsPrefAct = QAction('Preferences', '&Preferences...', 0, self)
        toolsPrefAct.setStatusTip('Set user preferences')
        toolsPrefAct.addTo(toolsMenu)
        self.connect(toolsPrefAct, SIGNAL('activated()'), self.toolsPref)

        toolsMenu.insertSeparator()

        toolsBkColorAct = QAction('Background Color', '&Background Color...', \
                                  0, self)
        toolsBkColorAct.setStatusTip('Set view background color')
        toolsBkColorAct.addTo(toolsMenu)
        self.connect(toolsBkColorAct, SIGNAL('activated()'), self.toolsBkColor)

        toolsTxtColorAct = QAction('Text Color', '&Text Color...', \
                                   0, self)
        toolsTxtColorAct.setStatusTip('Set view text color')
        toolsTxtColorAct.addTo(toolsMenu)
        self.connect(toolsTxtColorAct, SIGNAL('activated()'),\
                     self.toolsTxtColor)

        FlyMainWin.toolBar.addSeparator()
        helpMenu = QPopupMenu(self)
        self.menuBar().insertItem('&Help', helpMenu)

        helpReadMeAct = QAction('View ReadMe', QIconSet(QPixmap(help_xpm)), \
                             '&View ReadMe', 0, self)
        helpReadMeAct.setStatusTip('View the ReadMe file')
        helpReadMeAct.addTo(helpMenu)
        helpReadMeAct.addTo(FlyMainWin.toolBar)
        self.connect(helpReadMeAct, SIGNAL('activated()'), self.helpReadMe)

        helpAboutAct = QAction('About FlyWay', '&About FlyWay', 0, self)
        helpAboutAct.setStatusTip('About this program')
        helpAboutAct.addTo(helpMenu)
        self.connect(helpAboutAct, SIGNAL('activated()'), self.helpAbout)


class SearchDlg(QDialog):
    "Waypoint selection dialog after search returns multiple waypoints"
    filterNames = ('Include Airports', 'Include Navaids', 'Include Fixes')
    wpTypes = ('A', 'N', 'F')
    def __init__(self, wpList, searchStr, parent=None, name=None, \
                 modal=0, flags=0):
        QDialog.__init__(self, parent, name, modal, flags)
        self.wpList = wpList
        if sys.platform == 'win32':
            self.setCaption('Search Results  (PyQt)')
        else:
            self.setCaption('Search Results')
        self.setIcon(QPixmap(fly_xpm))
        topLayout = QVBoxLayout(self, 5)
        strLayout = QHBoxLayout(topLayout)

        self.searchEdit = QLineEdit(searchStr, self)
        strLayout.addWidget(self.searchEdit)
        self.connect(self.searchEdit, SIGNAL('textChanged(const QString&)'), \
                     self.enableSearch)
        self.connect(self.searchEdit, SIGNAL('returnPressed()'), \
                     self.repeatSearch)
        self.searchButton = QPushButton('Search', self)
        strLayout.addWidget(self.searchButton)
        self.connect(self.searchButton, SIGNAL('clicked()'), self.repeatSearch)
        self.searchButton.setEnabled(0)

        self.filters = QHButtonGroup('Filter Waypoints', self)
        topLayout.addWidget(self.filters)
        for name in SearchDlg.filterNames:
            button = QCheckBox(name, self.filters)
            button.setChecked(1)
        self.connect(self.filters, SIGNAL('clicked(int)'), self.filterList)
        self.enableFilters()

        self.listView = QListView(self)
        topLayout.addWidget(self.listView)
        self.listView.addColumn('#', 40)
        self.listView.addColumn('ID', 60)
        self.listView.addColumn('Description', 300)
        self.listView.setSorting(0)
        self.connect(self.listView, SIGNAL('doubleClicked(QListViewItem*)'), \
                     self, SLOT('accept()'))
        self.connect(self.listView, SIGNAL('returnPressed(QListViewItem*)'), \
                     self, SLOT('accept()'))
        self.filterList()

        ctrlLayout = QHBoxLayout(topLayout)
        cancelButton = QPushButton('Cancel', self)
        ctrlLayout.addWidget(cancelButton)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))
        okButton = QPushButton('OK', self)
        ctrlLayout.addWidget(okButton)
        self.connect(okButton, SIGNAL('clicked()'), self, SLOT('accept()'))

        self.listView.setFocus()

    def result(self):
        "Return selected waypoint or None"
        if self.listView.selectedItem():
            for wp in self.wpList:
                if wp.ident() == str(self.listView.selectedItem().text(1)) \
                   and wp.description() == str(self.listView.selectedItem().\
                   text(2)):
                    return wp
        return None

    def filterList(self):
        "Load list view with filtered waypoints"
        self.listView.clear()
        types = [SearchDlg.wpTypes[num] for num in range(3) if \
                 self.filters.find(num).isChecked()]
        num = 1
        for wp in self.wpList:
            if wp.wpTypeChar() in types:
                QListViewItem(self.listView, '%4d' % num, wp.ident(), \
                              wp.description())
                num += 1
        if self.listView.childCount():
            self.listView.setSelected(self.listView.firstChild(), 1)

    def enableFilters(self):
        "Set wp type filter buttons enabled based on wp contents"
        availTypes = [wp.wpTypeChar() for wp in self.wpList]
        for num in range(3):
            self.filters.find(num).setEnabled(SearchDlg.wpTypes[num] \
                                              in availTypes)

    def repeatSearch(self):
        "Do a new search based on text entry"
        text = str(self.searchEdit.text())
        if text:
            self.wpList = FlyMainWin.route.findWp(text)
            self.enableFilters()
            self.filterList()
            self.searchButton.setEnabled(0)
            self.listView.setFocus()

    def enableSearch(self):
        "Activate search button based on line edit change"
        self.searchButton.setEnabled( not self.searchEdit.text().isEmpty())


class TimeDlg(QDialog):
    "Depart time selection dialog with timezones & 12/24 hour selections"
    def __init__(self, parent=None, name=None, modal=0, flags=0):
        QDialog.__init__(self, parent, name, modal, flags)
        if sys.platform == 'win32':
            self.setCaption('Departure Time  (PyQt)')
        else:
            self.setCaption('Departure Time')
        self.setIcon(QPixmap(fly_xpm))
        defaultTime = Time(FlyMainWin.option.intData('DepartHour', 0, 23) \
                           + FlyMainWin.option.intData('DepartMinute', 0, 60) \
                           / 60.0)
        self.is12hrClock = FlyMainWin.option.boolData('12hrClock')
        self.localClock = FlyMainWin.option.boolData('LocalClock')

        topLayout = QVBoxLayout(self, 5)
        horizLayout = QHBoxLayout(topLayout)

        typeGroup = QVButtonGroup('Type', self)
        QRadioButton('Local Time, 12-Hour       ', typeGroup)
        QRadioButton('Local Time, 24-Hour       ', typeGroup)
        QRadioButton('Zulu Time, 24-Hour        ', typeGroup)
        if not self.localClock:
            typeGroup.setButton(2)
        elif self.is12hrClock:
            typeGroup.setButton(0)
        else:
            typeGroup.setButton(1)
        self.connect(typeGroup, SIGNAL('clicked(int)'), self.typeChange)
        horizLayout.addWidget(typeGroup)

        rightLayout = QVBoxLayout(horizLayout)
        self.curTimeLbl = QLabel(self)
        self.curTimeLbl.setFrameStyle(QFrame.Panel | QFrame.Sunken)
        self.curTimeLbl.setAlignment(Qt.AlignCenter)
        rightLayout.addWidget(self.curTimeLbl)
        self.setCurrentTime()

        editLayout = QHBoxLayout(rightLayout)
        self.hourBox = SpinBoxEx(0, 23, 1, 2, self)
        self.hourBox.setWrapping(1)
        editLayout.addWidget(self.hourBox, 0, Qt.AlignLeft)
        self.connect(self.hourBox, PYSIGNAL('spinChange'), self.spinHour)
        colon = QLabel(':', self)
        editLayout.addWidget(colon, 0, Qt.AlignLeft)
        self.minuteBox = SpinBoxEx(0, 59, 5, 2, self)
        self.minuteBox.setWrapping(1)
        editLayout.addWidget(self.minuteBox, 0, Qt.AlignLeft)
        self.connect(self.minuteBox, PYSIGNAL('wrapped'), self.wrapMinute)
        self.amBox = QComboBox(0, self)
        self.amBox.insertItem('AM')
        self.amBox.insertItem('PM')
        editLayout.addWidget(self.amBox)
        editLayout.addStretch(10)
        self.setDefaultTime(defaultTime)

        ctrlLayout = QHBoxLayout(topLayout)
        okButton = QPushButton('OK', self)
        ctrlLayout.addWidget(okButton)
        self.connect(okButton, SIGNAL('clicked()'), self, SLOT('accept()'))
        cancelButton = QPushButton('Cancel', self)
        ctrlLayout.addWidget(cancelButton)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))

    def setDefaultTime(self, time):
        "Update default time setting for types"
        self.hourBox.setMinValue(self.is12hrClock)
        self.hourBox.setMaxValue(self.is12hrClock and 12 or 23)
        self.hourBox.setValue(time.hourMinute(not self.is12hrClock)[0])
        self.minuteBox.setValue(time.hourMinute()[1])
        self.amBox.setCurrentItem(not time.isAM())
        if self.is12hrClock:
            self.amBox.show()
        else:
            self.amBox.hide()

    def setCurrentTime(self):
        "Set dialog current time string"
        text = Time().current(self.localClock).timeStr(not self.is12hrClock, \
                                                       self.is12hrClock)
        self.curTimeLbl.setText('Current Time:\n%s %s' % \
                                (text, self.localClock and 'Local' or 'Zulu'))

    def timeResult(self):
        "Return user time setting"
        hour = self.hourBox.value() + self.minuteBox.value() / 60.0
        if self.is12hrClock and self.amBox.currentItem == 1:
            hour += 12
        return Time(hour)

    def typeChange(self, id):
        "Change time style based on button group signal"
        timeSet = self.timeResult()
        if id < 2:   # local
            self.is12hrClock = id == 0
            if not self.localClock:
                timeSet.toLocal()
                self.localClock = 1
        elif self.localClock:   # GMT
            timeSet.toGMT()
            self.localClock = 0
            self.is12hrClock = 0
        self.setDefaultTime(timeSet)
        self.setCurrentTime()

    def spinHour(self, isUp, value):
        "Adjust AM & PM for hour spin change"
        if self.is12hrClock and (isUp and value == 12) \
           or (not isUp and value == 11):
            self.amBox.setCurrentItem(not self.amBox.currentItem())

    def wrapMinute(self, isUp):
        "Adjust hour on minute spin change"
        if isUp:
            self.hourBox.stepUp()
        else:
            self.hourBox.stepDown()

    def accept(self):
        "Set option from result on dialog accept signal"
        timeSet = self.timeResult().hourMinute()
        FlyMainWin.option.changeData('DepartHour', `timeSet[0]`, 0)
        FlyMainWin.option.changeData('DepartMinute', `timeSet[1]`, 0)
        FlyMainWin.option.changeData('12hrClock', \
                                     self.is12hrClock and 'yes' or 'no', 1)
        FlyMainWin.option.changeData('LocalClock', \
                                     self.localClock and 'yes' or 'no', 1)
        FlyMainWin.option.writeChanges()
        QDialog.accept(self)


def main():
    app = QApplication(sys.argv)
    win = FlyMainWin()
    app.setMainWidget(win)

    try:
        opts, args = getopt.getopt(sys.argv[1:], '')
    except getopt.GetoptError:
        print 'Bad command line option'
        sys.exit(2)
    if args:
        win.openFile(args[0])
    win.show()
    app.exec_loop()

if __name__ == '__main__':
    main()
