#!/usr/bin/env python

#****************************************************************************
# recentfiles.py, class to handle recent file lists in menus
#
# Copyright (C) 2002, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#*****************************************************************************

import os.path
from option import Option
from qt import *

class RecentFiles(QObject):
    "Tracks & loads menus with recently opened files"
    def __init__(self, option, numEntries, maxEntryLen, parent=None, \
                 name=None):
        QObject.__init__(self, parent, name)
        self.option = option
        self.numEntries = numEntries
        self.maxEntryLen = maxEntryLen
        self.menu = None
        self.firstId = 0
        self.firstIndex = 0

    def initMenu(self, menu, firstId):
        "load the entries into the menu"
        self.menu = menu
        self.firstId = firstId
        self.firstIndex = self.menu.count()
        for i in range(self.numEntries):
            text = self.menuText(i)
            if text:
                self.menu.insertItem(text, self.firstId + i)
        self.connect(self.menu, SIGNAL('activated(int)'), self.openItem)

    def addEntry(self, fileName):
        "Add a new filename to the start of the list"
        newPath = os.path.abspath(fileName)
        curPath = newPath
        for i in range(self.numEntries):
            tmpPath = self.option.strData(self.optionTitle(i), 1)
            self.option.changeData(self.optionTitle(i), curPath, 1)
            curPath = tmpPath
            if curPath == newPath:
                break           # eliminate duplicates
        self.option.writeChanges()
        self.updateMenu()

    def removeEntry(self, filePath):
        "Remove a filename from the list"
        for i in range(self.numEntries):
            if filePath == self.option.strData(self.optionTitle(i), 1):
                self.option.changeData(self.optionTitle(i), '', 1)
                self.option.writeChanges()
                self.updateMenu()
                return

    def openItem(self, id):
        "Signal to main window to open an item"
        if self.firstId <= id < self.firstId + self.numEntries:
            self.emit(PYSIGNAL('askOpen'),(self.option.strData(\
                self.optionTitle(id - self.firstId), 1),))

    def updateMenu(self):
        "Refresh menu items"
        for i in range(self.numEntries):
            num = self.menu.indexOf(self.firstId + i)
            text = self.menuText(i)
            if text:
                if num >= 0:
                    self.menu.changeItem(self.firstId + i, text)
                else:
                    self.menu.insertItem(text, self.firstId + i, \
                                         self.firstIndex + i)
            else:
                if num >= 0:
                    self.menu.removeItem(self.firstId + i)

    def optionTitle(self, num):
        "Return option key"
        return 'RecentFile' + `num + 1`

    def menuText(self, num):
        "Return abbreviated path name"
        path = self.option.strData(self.optionTitle(num), 1)
        if len(path) > self.maxEntryLen:
            pos = path.find('/', len(path) - (self.maxEntryLen - 3))
            if pos < 0:
                pos = len(path) - (self.maxEntryLen - 3)
            path = '...' + path[pos:]
        if path:
            path = '&%d %s' % (num+1, path)
        return path
